# This code is part of KQCircuits
# Copyright (C) 2021 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).


import ast
import re

from kqcircuits.pya_resolver import pya
from kqcircuits.elements.waveguide_composite import Node, WaveguideComposite
from kqcircuits.util.library_helper import load_libraries, element_by_class_name


def get_nodes_near_position(top_cell, position, box_size=10, require_gui_editing_enabled=True):
    """Find all WaveguideComposite nodes near a specified position.

    Considers only waveguides that are a direct child of the specified ``top_cell``.

    Args:
        top_cell: cell in which to search for WaveguideComposite instances
        position: pya.DPoint position where to search for nodes
        box_size: capture distance in x,y away from ``position`` where the node can be
        require_gui_editing_enabled: if True, only instances with ``enable_gui_editing==True`` are considered.

    Returns:
        a list of tuples ``(instance, node, node_index)`` where ``instance`` is the WaveguideComposite Instance,
        ``node`` is the ``Node`` object that ``node_index`` is the index of ``node`` in the ``nodes`` parameter of
        the waveguide.
    """
    box = pya.DBox(pya.DPoint(-box_size, -box_size), pya.DPoint(box_size, box_size)).moved(pya.DVector(position))

    found_nodes = []

    for inst in top_cell.each_inst():
        if inst.is_pcell() and isinstance(inst.pcell_declaration(), WaveguideComposite):
            if (not require_gui_editing_enabled) or inst.pcell_parameter('enable_gui_editing'):
                dtrans = inst.dcplx_trans

                nodes = Node.nodes_from_string(inst.pcell_parameter("nodes"))
                for i, node in enumerate(nodes):
                    node_position = dtrans * node.position
                    if box.contains(node_position):
                        found_nodes.append((inst, node, i))
    return found_nodes


def node_to_text(node):
    """Convert ``Node`` object to text fields that can be used for GUI editing.

    The inverse of this function is ``node_from_text``.

    Args:
        node: Node to convert

    Returns:
        tuple of strings ``(x, y, element, inst_name, angle, length_before, length_increment, align, parameters)``
    """

    x = str(node.position.x)
    y = str(node.position.y)
    inst_name = node.inst_name if node.inst_name is not None else ""
    angle = str(node.angle) if node.angle is not None else ""
    length_before = str(node.length_before) if node.length_before is not None else ""
    length_increment = str(node.length_increment) if node.length_increment is not None else ""
    element = node.element.__name__ if node.element is not None else ""
    align = ",".join(node.align) if node.align is not None else ""
    parameters = "\n".join([f'{key}={repr(value)}' for key, value in node.params.items()])

    return (x, y, element, inst_name, angle, length_before, length_increment, align, parameters)


def node_from_text(x, y, element, inst_name, angle, length_before, length_increment, align, parameters):
    """Create Node from text inputs.

    This is the inverse of ``node_to_text``.

    For all arguments except x and y, an empty string is treated as default value. Spaces are stripped from the inputs.
    For ``parameters``, the values will be parsed by ``ast.literal_eval``, which accepts most standard python literals.

    Args:
        x (str): x position, will be converted to float
        y (str): y position, will be converted to float
        element (str): class name of the element, must exist in the `kqcircuits.elements` namespace
        inst_name (str): instance name to use for the element
        angle (str): angle of the node, will be converted to float
        length_before (str): length before this node, will be converted to float
        length_increment (str): length increment produced by meander before this node, will be converted to float
        align (str): input and output refpoint to use for aligning the element, separated by a comma
        parameters (str): multiline string, where each line contains a parameter=value pair.

    Returns: Node

    Raises:
        ValueError: if an input value cannot be converted to the correct type.
    """
    x = float(x)
    y = float(y)

    params = {}
    if inst_name != "":
        params['inst_name'] = inst_name.strip()
    if angle != "":
        params['angle'] = float(angle)
    if length_before != "":
        params['length_before'] = float(length_before)
    if length_increment != "":
        params['length_increment'] = float(length_increment)
    element = element.strip()
    if element == "":
        element = None
    if align != "":
        params['align'] = tuple(s.strip() for s in align.split(','))

    param_lines = parameters.split('\n')
    for param_line in param_lines:
        param_line = param_line.strip()
        if param_line != "":
            m = re.fullmatch(r'([a-zA-Z0-9_]+)\s*=\s*(.*)', param_line)
            if not m:
                raise ValueError('Element parameters should be one key=value pair on each line')

            key = m.groups()[0]
            value = ast.literal_eval(m.groups()[1])
            params[key] = value

    return Node.deserialize((x, y, element, params))


def replace_node(waveguide_instance, node_index, node):
    """Replace a Node in a WaveguideComposite by index.

    Args:
        waveguide_instance: Instance of the waveguide
        node_index: (int) index of the node to replace
        node: new Node
    """
    nodes = Node.nodes_from_string(waveguide_instance.pcell_parameter('nodes'))
    nodes[node_index] = node
    waveguide_instance.change_pcell_parameter('nodes', [str(node) for node in nodes])

    # Update gui_path and gui_path_shadow to reflect possible position changes in the node
    if waveguide_instance.pcell_parameter("enable_gui_editing"):
        path = pya.DPath([node.position for node in nodes], 1)
        # Round to database units since the KLayout partial tool also works in database units
        dbu = waveguide_instance.layout().dbu
        path = path.to_itype(dbu).to_dtype(dbu)
        waveguide_instance.change_pcell_parameter("gui_path", path)
        waveguide_instance.change_pcell_parameter("gui_path_shadow", path)


def get_all_node_elements():
    """Returns all class names from PCells in the Element library which can be used in WaveguideComposite nodes

    Returns:
        List of class names (str)
    """
    valid_elements = ['Airbridge']

    layout = load_libraries(path='elements')['Element Library'].layout()
    for pcell_id in layout.pcell_ids():
        pcell = layout.pcell_declaration(pcell_id)
        valid_elements.append(pcell.__class__.__name__)

    return valid_elements


def get_valid_node_elements():
    """Returns a list of all element class names which would be, at least in principle, usable as ```Node.element```.

    An element is considered valid if it has at least two refpoint pairs ```X``` and ```X_corner```, for any value of
    x.

    Note: this function creates each element with default parameter values. Since this is generally slow and clumsy,
    this function is not used at startup. Instead, a curated list ```node_editor_valid_elements``` is kept in
    ```kqcircuits.defaults```.

    Returns:
        List of class names (str)
    """
    layout = pya.Layout()

    all_elements = get_all_node_elements()
    valid_elements = []

    for classname in all_elements:
        element = element_by_class_name(classname)
        if element is not None:
            cell, refp = element.create_with_refpoints(layout)

            refpoints_with_corner = [name.rstrip('_corner') for name in refp
                                     if name.endswith('_corner') and name.rstrip('_corner') in refp]

            layout.delete_cell(cell.cell_index())

            if len(refpoints_with_corner) > 1:
                valid_elements.append(classname)
        else:
            valid_elements.append(classname)

    return valid_elements



# This code is part of KQCircuits
# Copyright (C) 2021 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).

import importlib
import networkx as nx
from kqcircuits.defaults import default_netlist_ignore_connections

spec = importlib.util.find_spec("matplotlib")
matplotlib_exists = spec is not None
if matplotlib_exists:
    from matplotlib import pyplot as plt


def network_as_graph(network):
    """
    Import KQC netlist as networkx graph.

    Each element is added as a node, identified by the subcircuit id used in the netlist.
    For each node, the data dictionary contains the following items:

    * cell_name: identifies the PCell, for example ``Waveguide Coplanar$1``.
        Multiple nodes can point to the same PCell, if it has identical parameters.
    * cell_type: The PCell full name, for example ``Waveguide Coplanar``
    * location: A list [x, y] specifying the element coordinates in um.
    * instance_name: The instance name of the PCell instance, may be an empty string
    * name: A unique name for this instance. Equal to instance_name if that exists, made
        unique by appending a numbered suffix if needed. If the instance name is empty,
        a string containing the subcircuit_id.
    * properties: A dictionary containing the PCell properties associated with this instance

    Args:
        network: dictionary of netlist data obtained by loading the netlist json file

    Returns. Networkx Graph
    """
    edges = []
    used_subcircuit_ids = set()
    used_names = set()

    # Add all edges from the netlist
    for net in network["nets"].values():
        if len(net) >= 2:
            for i,net_i in enumerate(net):
                for net_j in net[i+1:]:
                    reasons_to_ignore_connections = [(a,b) for a,b in default_netlist_ignore_connections
                        if (net_i["pin"] == a and net_j["pin"] == b) or (net_i["pin"] == b and net_j["pin"] == a)]
                    if len(reasons_to_ignore_connections) > 0:
                        continue
                    edges.append([net_i["subcircuit_id"], net_j["subcircuit_id"]])
                    used_subcircuit_ids.add(net_i["subcircuit_id"])
                    used_subcircuit_ids.add(net_j["subcircuit_id"])
    graph = nx.Graph()
    graph.add_edges_from(edges)

    # Add data to the nodes
    for subcircuit_id in used_subcircuit_ids:
        subcircuit = network["subcircuits"][str(subcircuit_id)]
        graph.nodes[subcircuit_id]["cell_name"] = subcircuit["cell_name"]
        graph.nodes[subcircuit_id]["cell_type"] = subcircuit["cell_name"].split('$')[0].replace('*', ' ')
        graph.nodes[subcircuit_id]["location"] = [  subcircuit["subcircuit_location"]["x"],
                                                    subcircuit["subcircuit_location"]["y"]]
        if "instance_name" in subcircuit and subcircuit["instance_name"] is not None:
            instance_name = subcircuit["instance_name"]
        else:
            instance_name = ""
        base_name = instance_name if instance_name != "" else str(subcircuit_id)

        # Define a unique name by suffixing with a number if needed
        name = base_name
        i = 0
        while name in used_names:
            i += 1
            name = f"{base_name}_{i}"
        used_names.add(base_name)

        graph.nodes[subcircuit_id]["instance_name"] = instance_name
        graph.nodes[subcircuit_id]["name"] = name
        graph.nodes[subcircuit_id]["properties"] = subcircuit.get("properties", {})

    return graph


def draw_graph(graph, with_labels=True, with_position=True, figsize=(8, 8), export_path=None):
    """
    Draw a netlist graph

    Args:
        graph: Networkx Graph with data structures as loaded by ``network_as_graph``
        with_labels: if True, the unique ``name`` of each node will be shown as label
        with_position: if True, the nodes will be positioned as they are physically located on the chip.
            If False, a spring layout will be used to position the nodes.
        figsize: Figure size to pass to matplotlib, default (8, 8)
        export_path: Path to export image to, or None to show interactive plot

    """
    if not matplotlib_exists:
        return
    plt.figure(figsize=figsize)
    if with_position:
        pos = {node: data["location"] for node, data in graph.nodes(data=True)}
    else:
        pos = nx.spring_layout(graph)
    labels = {node: data["name"] for node, data in graph.nodes(data=True)}
    nx.draw(graph, pos=pos, labels=labels, with_labels=with_labels)
    if export_path is not None:
        plt.savefig(export_path)
    else:
        plt.show()



# This code is part of KQCircuits
# Copyright (C) 2023 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).


from kqcircuits.pya_resolver import pya
from kqcircuits.util.parameters import Param, pdt

from kqcircuits.test_structures.test_structure import TestStructure


class StripesTestIncreasingWidth(TestStructure):
    """PCell declaration for optical lithography test stripes.

    Contains a given number of lines in `base_metal_gap_wo_grid`-layer with given length and spacing, and increasing
    width. There is also a text in `base_metal_gap_wo_grid`-layer next to the lines, which shows the line width.
    """

    num_stripes = Param(pdt.TypeInt, "Number of stripes", 20)
    min_stripe_width = Param(pdt.TypeDouble, "Beginning width of the stripe sequence", 1, unit="μm")
    stripe_step = Param(pdt.TypeDouble, "Stripe widening increment in the stripe sequence", 1, unit="μm")
    stripe_length = Param(pdt.TypeDouble, "Length of the stripes", 100, unit="μm")
    stripe_spacing = Param(pdt.TypeDouble, "Spacing between the stripes", 1, unit="μm")

    def build(self):
        layer_base_metal = self.get_layer("base_metal_gap_wo_grid")

        width = float(self.min_stripe_width)

        loc = 0
        for _ in range(self.num_stripes):
            stripe = pya.DBox(0, 0, width, self.stripe_length)
            trans = pya.DTrans(loc, 0)
            self.cell.shapes(layer_base_metal).insert(trans * stripe)
            loc += width + self.stripe_spacing
            width += float(self.stripe_step)

        spacing_str = int(self.stripe_spacing)
        text_cell = self.layout.create_cell("TEXT", "Basic", {
            "layer": self.face()["base_metal_gap_wo_grid"],
            "text": "{}".format(spacing_str),
            "mag": 40,
        })
        text_x = float(loc + self.min_stripe_width)
        text_y = self.stripe_length / 2
        self.insert_cell(text_cell, pya.DTrans(text_x, text_y))



# This code is part of KQCircuits
# Copyright (C) 2022 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).


import abc

import logging

from kqcircuits.elements.element import Element
from kqcircuits.pya_resolver import pya
from kqcircuits.simulations.simulation import Simulation, get_simulation_layer_by_name
from kqcircuits.util.parameters import Param, pdt, add_parameters_from


@add_parameters_from(Simulation, 'name', 'box', 'extra_json_data')
class CrossSectionSimulation:
    """Class for co-planar waveguide cross-section simulations.

    This class is intended to be subclassed by a specific simulation implementation;  The subclass should implement the
    method 'build' which defines the simulation geometry and material properties.

    Layer names in cross-section geometry don't need to obey KQC default layers. The layers use following name coding:
    - Layer name 'vacuum' is reserved for vacuum material. The vacuum layer can be left empty; then the simulation
    export fills the empty space inside the 'box' with vacuum.
    - All layers that include word 'signal' are considered as signal metals. Different signal layers are considered as
    separate signals so that the result matrices has row and column for each signal layer. The order of signals is the
    same as where they are introduced by calling 'get_sim_layer' function.
    - All layers that include word 'ground'  are considered as ground metal.
    - Any other layer is considered as dielectric. The permittivity can be set using the 'set_permittivity' function.
    """

    LIBRARY_NAME = None  # This is needed by some methods inherited from Element.

    london_penetration_depth = Param(
        pdt.TypeDouble, "London penetration depth of metals", 0.0, unit='m',
        docstring="London penetration depth is implemented for one signal simulation only"
    )
    xsection_source_class = Param(
        pdt.TypeNone, "Simulation class XSection tool was used on", None, docstring=
        "Class from which the simulation was generated from using the XSection tool. Used to get the correct schema."
    )

    def __init__(self, layout, **kwargs):
        """Initialize a CrossSectionSimulation.

        The initializer parses parameters, creates a top cell, and then calls `self.build` to create
        the simulation geometry.

        Args:
            layout: the layout on which to create the simulation

        Keyword arguments:
            `**kwargs`:
                Any parameter can be passed as a keyword argument.
        """
        if layout is None or not isinstance(layout, pya.Layout):
            error_text = "Cannot create simulation with invalid or nil layout."
            error = ValueError(error_text)
            logging.error(error_text)
            raise error

        self.layout = layout
        schema = type(self).get_schema()

        # Apply kwargs or default value
        for parameter, item in schema.items():
            if parameter in kwargs:
                setattr(self, parameter, kwargs[parameter])
            else:
                setattr(self, parameter, item.default)

        self.cell = kwargs['cell'] if 'cell' in kwargs else layout.create_cell(self.name)

        self.layer_dict = dict()
        self.permittivity_dict = dict()
        self.units = kwargs.get('units', 'um')
        self.build()

    # Inherit specific methods from Element
    get_schema = classmethod(Element.get_schema.__func__)

    @abc.abstractmethod
    def build(self):
        """Build simulation geometry.

        This method is to be overridden, and the overriding method should create the geometry to be simulated.
        """
        return

    def register_cell_layers_as_sim_layers(self):
        """Takes all layers that contain any geometry and registers them as simulation layers

        This method resets the internal simulation layer dictionary.
        """
        self.layer_dict = dict()
        self.permittivity_dict = dict()
        for l in self.layout.layer_infos():
            if len(list(self.cell.each_shape(self.layout.layer(l)))) > 0:
                self.layer_dict[l.name] = l

    def get_sim_layer(self, layer_name):
        """Returns layer of given name. If layer doesn't exist, a new layer is created."""
        if layer_name not in self.layer_dict:
            layer_info = None
            for l in self.layout.layer_infos():
                if l.datatype == 0 and l.name == layer_name:
                    # If there is a layer with layer_name in layout (used by other cell), reuse that
                    layer_info = l
            if layer_info is None:
                layer_info = get_simulation_layer_by_name(layer_name)
            self.layer_dict[layer_name] = layer_info
        return self.layout.layer(self.layer_dict[layer_name])

    def set_permittivity(self, layer_name, permittivity):
        """Sets permittivity for layer of given name."""
        self.permittivity_dict[layer_name] = permittivity

    def get_parameters(self):
        """Return dictionary with all parameters and their values."""
        return {
            param: getattr(self, param)
            for param in (self.xsection_source_class or type(self)).get_schema()
        }

    def get_simulation_data(self):
        """Return the simulation data in dictionary form.

        Returns:
            dictionary of relevant parameters for simulation
        """
        simulation_data = {
            'gds_file': self.name + '.gds',
            'units': self.units,
            'box': self.box,
            "london_penetration_depth": self.london_penetration_depth,
            **{'{}_permittivity'.format(k): v for k, v in self.permittivity_dict.items()},
            'parameters': self.get_parameters(),
        }
        return simulation_data



# This code is part of KQCircuits
# Copyright (C) 2021 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).

import copy
import os
import logging
from sys import argv
from time import perf_counter
from inspect import isclass
from multiprocessing import Pool
from pathlib import Path

from tqdm import tqdm

from kqcircuits.chips.chip import Chip
from kqcircuits.masks.multi_face_mask_layout import MultiFaceMaskLayout
from kqcircuits.util.log_router import route_log
from kqcircuits.pya_resolver import pya, is_standalone_session
from kqcircuits.defaults import default_bar_format, TMP_PATH, default_face_id
from kqcircuits.masks.mask_export import export_chip, export_mask_set
from kqcircuits.masks.mask_layout import MaskLayout
from kqcircuits.klayout_view import KLayoutView


class MaskSet:
    """Class representing a set of masks for different chip faces.

    A mask set consists of one or more MaskLayouts, each of which is for a certain face.

    To create a mask, add mask layouts to the mask set using add_mask_layout() and add chips to these mask layouts using
    add_chip(). These functions also export some files for each chip. Then call build() to create the
    cell hierarchy of the entire mask, and finally export mask files by calling export().

    Chips are created in parallel in separate processes but the user may choose to use a ``-d`` switch on
    the command line for debugging with a single process. It is also possible to manually limit the number of
    concurrently used CPUs for resource management purposes with the ``-c 4`` switch (to 4 in this example).

    Example:
        mask = MaskSet(...)
        mask.add_mask_layout(...)
        mask.add_mask_layout(...)
        mask.add_chip(...)
        mask.build()
        mask.export()

    Attributes:
        layout: pya.Layout of this mask set
        name: Name of the mask set
        version: Version of the mask set
        with_grid: Boolean determining if ground grid is generated
        export_drc: Boolean determining if DRC report is exported
        chips_map_legend: Dictionary where keys are chip names, values are chip cells
        mask_layouts: list of MaskLayout objects in this mask set
        mask_export_layers: list of names of the layers which are exported for each MaskLayout
        used_chips: similar to chips_map_legend, but only includes chips which are actually used in mask layouts
        export_path: The folder for mask files will be generated under this. TMP_PATH by default.
    """

    def __init__(self, view=None, name="MaskSet", version=1, with_grid=False, export_drc=False,
                 mask_export_layers=None, export_path=TMP_PATH):

        self._time = {"INIT": perf_counter(), "ADD_CHIPS": 0,  "BUILD": 0, 'EXPORT': 0, 'END': 0}

        if view is None:
            self.view = KLayoutView()
        self.layout = self.view.layout

        self.name = name
        self.version = version
        self.with_grid = with_grid
        self.export_drc = export_drc
        self.chips_map_legend = {}
        self.mask_layouts = []
        self.mask_export_layers = mask_export_layers if mask_export_layers is not None else []
        self.used_chips = {}
        self._extra_params = {}
        self._mask_set_dir = Path(export_path)/f"{name}_v{version}"

        self._mask_set_dir.mkdir(parents=True, exist_ok=True)

        self._extra_params["enable_debug"] = '-d' in argv
        self._single_process = self._extra_params["enable_debug"] or not is_standalone_session()

        self._extra_params["mock_chips"] = '-m' in argv
        self._extra_params["skip_extras"] = '-s' in argv

        self._cpu_override = 0
        if '-c' in argv and len(argv) > argv.index('-c') + 1:
            self._cpu_override = int(argv[argv.index('-c') + 1])

    def add_mask_layout(self, chips_map, face_id=default_face_id, mask_layout_type=MaskLayout, **kwargs):
        """Creates a mask layout from chips_map and adds it to self.mask_layouts.

        Args:
            chips_map: List of lists (2D-array) of strings, each string is a chip name (or --- for no chip)
            face_id: face_id of the mask layout
            mask_layout_type: type of the mask layout (MaskLayout or a child class of it)
            kwargs: keyword arguments passed to the mask layout

        Returns:
            the created mask layout
        """
        if ("mask_export_layers" not in kwargs) and self.mask_export_layers:
            kwargs["mask_export_layers"] = self.mask_export_layers

        mask_layout = mask_layout_type(self.layout, self.name, self.version, self.with_grid, chips_map, face_id,
                                       **kwargs)
        self.mask_layouts.append(mask_layout)
        return mask_layout

    def add_multi_face_mask_layout(self, face_ids, chips_map=None, extra_face_params=None, mask_layout_type=MaskLayout,
                                   **kwargs):
        """Create a multi face mask layout, which can be used to make masks with matching chip maps on multiple faces.

        A ``MaskLayout`` is created of each face in ``face_ids``. By default, the individual mask layouts all have
        identical parameters, but parameters can be overwritten for a single face id through ``extra_face_params``.

        By default, ``bbox_face_ids`` is set to ``face_ids`` for all mask layouts.

        Args:
            face_ids: list of face ids to include
            chips_map: Chips map to use, or None to use an empty chips map.
            extra_face_params: a dictionary of ``{face_id: extra_kwargs}``, where ``extra_kwargs`` is a dictionary of
                 keyword arguments to apply only to the mask layout for ``face_id``.
            mask_layout_type: optional subclass of MaskLayout to use
            kwargs: any keyword arguments are passed to all containing mask layouts.

        Returns: a ``MultiFaceMaskLayout`` instance
        """
        if ("mask_export_layers" not in kwargs) and self.mask_export_layers:
            kwargs["mask_export_layers"] = self.mask_export_layers

        mfml = MultiFaceMaskLayout(self.layout, self.name, self.version, self.with_grid, face_ids,
                                   chips_map, extra_face_params, mask_layout_type, **kwargs)
        for face_id in mfml.face_ids:
            self.mask_layouts.append(mfml.mask_layouts[face_id])
        return mfml

    def add_chip(self, chips, variant_name=None, cpus=None, **parameters):
        """Adds a chip (or list of chips) with parameters to self.chips_map_legend and exports the files for each chip.

        Note the complex polymorphism used here: ``chips`` is either a single chip class or a list of ``(chip, variant,
        parameters)`` tuples. In the latter case the rest of the arguments (except ``cpus``) are ignored. Also,
        ``chips`` (or the individual chip part of tuples) may be a simple file name to load a static .oas file instead.

        The chip's parameters dictionary may also contain an ``alt_netlists`` dictionary to specify alternative ways of
        generating netlists. See ``export_cell_netlist()`` or the ``quick_demo.py`` mask for further information.

        Args:
            chip: A chip class. Or a list of tuples, like ``[(QualityFactor, "QDG", parameters),...]``,
                  parameters are optional.
            variant_name: Name for specific variant, the same as in the mask layout.
            cpus: Number of parallel processes to use for chip generation. By default uses ``os.cpu_count()``
                  or the number of chips, whichever is smaller.
            **parameters: Any parameters passed to the a single chip PCell.
        """
        self._time['ADD_CHIPS'] = perf_counter()

        if not isinstance(chips, list):  # only one chip
            cpus = 1
            chips = [(chips, variant_name, parameters)]

        if cpus is None:
            cpus = min(len(chips), os.cpu_count())
        if self._cpu_override > 0:
            cpus = self._cpu_override

        # Pool.map() needs all arguments packed into a single list
        xargs = (self.name, self.with_grid, self._mask_set_dir, self.export_drc, self._extra_params)
        chip_args = ((chip, xargs) for chip in chips)

        file_names = []
        if cpus == 1 or self._single_process:
            file_names += map(self._create_chip, chip_args)
        else:
            print(f"Building chip variants in parallel using {cpus} processes...")
            with Pool(cpus) as pool:
                file_names += pool.map(self._create_chip, chip_args)

        # import chip cells exported by the parallel processes into the mask
        for variant, file_name in tqdm(file_names, desc='Add chips into mask', bar_format=default_bar_format):
            self._load_chip_into_mask(file_name, variant)

    @staticmethod
    def _create_chip(chip_arg):
        """Create chip, possibly in a separate process."""

        chip, xargs = chip_arg
        name, with_grid, _mask_set_dir, export_drc, _extra_params = xargs
        chip_class, variant_name, *chip_params = chip
        chip_params = chip_params[0] if chip_params else {}
        alt_netlists = chip_params.pop("alt_netlists", None)

        chip_path = _mask_set_dir/"Chips"/f"{variant_name}"
        chip_path.mkdir(parents=True, exist_ok=True)

        logging.basicConfig(level=logging.DEBUG, force=True)  # this level is NOT actually used
        route_log(filename=chip_path/f"{variant_name}.log", stdout=_extra_params["enable_debug"])

        mock_chips = _extra_params['mock_chips']
        skip_extras = _extra_params['skip_extras']

        view = KLayoutView()
        layout = view.layout

        if isclass(chip_class):
            params = {
                'name_chip': variant_name,
                'name_mask': name,
                'with_grid': with_grid,
                'merge_base_metal_gap': True,
                'display_name': variant_name,
                'name_copy': None,
            }
            if mock_chips:
                mock_params = chip_class().pcell_params_by_name(Chip, **params)
                if chip_params:
                    # Pass through parameters only if they exist in Chip
                    mock_params.update({k: v for k, v in chip_params.items() if k in mock_params})
                mock_params.update({
                    'with_grid': False,
                    'with_gnd_bumps': False,
                    'with_gnd_tsvs': False,
                })
                cell = Chip.create(layout, **mock_params)
            else:
                if chip_params:
                    params.update(chip_params)
                cell = chip_class.create(layout, **params)
        else:  # it's a file name, load it
            load_opts = pya.LoadLayoutOptions()
            if hasattr(pya.LoadLayoutOptions, "CellConflictResolution"):
                load_opts.cell_conflict_resolution = pya.LoadLayoutOptions.CellConflictResolution.RenameCell
            layout.read(chip_class, load_opts)
            cell = layout.top_cells()[-1]

        export_chip(cell, variant_name, chip_path, layout, export_drc, alt_netlists, skip_extras)
        view.close()

        return variant_name, str(chip_path / f"{variant_name}.oas")

    def build(self, remove_guiding_shapes=True):
        """Builds the mask set.

        Creates cells for the mask based on self.mask_layouts and self.chips_map_legend. Optionally removes
        guiding shapes from the layout. Populates self.used_chips with the chips used in the mask layouts.

        Args:
            remove_guiding_shapes (Boolean): determines if the guiding shapes are removed

        """
        self._time['BUILD'] = perf_counter()
        # build mask layouts (without chip copy labels)
        for mask_layout in self.mask_layouts:
            # include face_id in mask_layout.name only for multi-face masks
            if len(self.mask_layouts) > 1:
                mask_layout.name += "-" + mask_layout.face_id
            mask_layout.build(self.chips_map_legend)

        chip_copy_label_layers = [
            "base_metal_gap",
            "base_metal_gap_wo_grid",
            "base_metal_gap_for_EBL"
        ]

        # Insert submask cells to different cell instances, so that these cells can have different chip labels even if
        # the original submask cells are identical. Also copy the MaskLayout objects of identical submasks into separate
        # MaskLayout objects with different `extra_id` so that mask export can use that information.
        mask_layouts_to_remove = set()
        submask_layouts_with_exported_layers = set()
        submask_layouts = []
        for mask_layout in self.mask_layouts:
            for i, (sm_layout, sm_pos) in enumerate(mask_layout.submasks):
                new_sm_layout = copy.copy(sm_layout)
                new_sm_layout.extra_id = f" s{i+1}"
                old_top_cell = new_sm_layout.top_cell
                new_sm_layout.top_cell = self.layout.create_cell(f"{new_sm_layout.name}{new_sm_layout.extra_id}")
                new_sm_layout.top_cell.insert(pya.DCellInstArray(old_top_cell.cell_index(), pya.DTrans()))
                mask_layout.top_cell.insert(
                    pya.DCellInstArray(new_sm_layout.top_cell.cell_index(),
                                       pya.DTrans(sm_pos - sm_layout.wafer_center + mask_layout.wafer_center))
                )
                mask_layout.submasks[i] = (new_sm_layout, sm_pos)
                submask_layouts.append(new_sm_layout)
                mask_layouts_to_remove.add(sm_layout)
                # Make sure that layers are only exported once if there are multiple identical submasks
                if sm_layout in submask_layouts_with_exported_layers:
                    # only export layers where chip copy labels are since they are different even for identical submasks
                    new_sm_layout.mask_export_layers = \
                        [layer for layer in chip_copy_label_layers if layer in new_sm_layout.mask_export_layers]
                else:
                    submask_layouts_with_exported_layers.add(sm_layout)
        self.mask_layouts = submask_layouts + [ml for ml in self.mask_layouts if ml not in mask_layouts_to_remove]

        # add chip copy labels for every mask layout
        for mask_layout in tqdm(self.mask_layouts, desc='Adding chip copy labels', bar_format=default_bar_format):

            labels_cell = mask_layout.layout.create_cell("ChipLabels")
            mask_layout.top_cell.insert(pya.DCellInstArray(labels_cell.cell_index(), pya.DTrans(pya.DVector(0, 0))))

            if mask_layout not in submask_layouts:
                mask_layout.insert_chip_copy_labels(labels_cell, chip_copy_label_layers)
                # remove "$1" or similar unnecessary postfix from cell name
                mask_layout.top_cell.name = f"{mask_layout.name}"

        # populate used_chips with chips which exist in some mask_layout
        for chip_name, cell in self.chips_map_legend.items():
            for mask_layout in self.mask_layouts:
                if mask_layout.chip_counts[chip_name]:
                    self.used_chips[chip_name] = cell
                    break

        # remove the guiding shapes, like chip boxes and waveguide paths
        if remove_guiding_shapes and self.layout.is_valid_layer(self.layout.guiding_shape_layer()):
            self.layout.delete_layer(self.layout.guiding_shape_layer())

    def export(self):
        """Exports designs, bitmaps and documentation of this mask set.

        Assumes that self.build() has been called before.
        """
        self._time['EXPORT'] = perf_counter()

        print("Exporting mask set...")
        export_mask_set(self, self._extra_params["skip_extras"])

        self._time['END'] = perf_counter()

        def tdiff(a, b):  # get elapsed time from "a" to "b"
            return f'{self._time[b] - self._time[a]:.1f}s' if self._time[a] and self._time[b] else 'n/a'

        print(f"Runtime: {tdiff('INIT', 'END')} (add chips: {tdiff('ADD_CHIPS', 'BUILD')}, "
              f"build: {tdiff('BUILD', 'EXPORT')}, export: {tdiff('EXPORT', 'END')})")

    @staticmethod
    def chips_map_from_box_map(box_map, mask_map):
        """Returns the chips_map created from box_map and mask_map.

        Given NxN box map and MxM mask_map, creates chips_map of size MNxMN. So each element of mask map is "replaced"
        by a box in the box map. Assumes that box_map and mask_map are square.

        Args:
            box_map: dictionary where keys are strings identifying the box type, and values are 2D arrays (lists of
                lists) where each element is a string identifying the chip type
            mask_map: 2D array (list of lists), where each element is a string identifying the box type

        """
        num_box_map_rows = len(list(box_map.values())[0])
        num_mask_map_rows = len(mask_map)
        num_chip_rows = num_box_map_rows*num_mask_map_rows

        chips_map = [["" for _ in range(num_chip_rows)] for _ in range(num_chip_rows)]
        for (k, box_row) in enumerate(mask_map):
            for (l, box) in enumerate(box_row):
                if box in box_map:
                    for (i, row) in enumerate(box_map[box]):
                        for (j, slot) in enumerate(row):
                            chips_map[k*num_box_map_rows + i][l*num_box_map_rows + j] = slot

        return chips_map

    def _load_chip_into_mask(self, file_name, variant_name):
        """Loads a chip from file_name to self.layout and adds it into self.chips_map_legend["variant_name"]"""
        load_opts = pya.LoadLayoutOptions()
        load_opts.cell_conflict_resolution = pya.LoadLayoutOptions.CellConflictResolution.RenameCell
        self.layout.read(file_name, load_opts)
        self.chips_map_legend.update({variant_name: self.layout.top_cells()[-1]})



# This code is part of KQCircuits
# Copyright (C) 2023 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).

import math

from kqcircuits.elements.element import Element
from kqcircuits.pya_resolver import pya
from kqcircuits.qubits.qubit import Qubit
from kqcircuits.util.geometry_helper import circle_polygon, arc_points
from kqcircuits.util.parameters import Param, pdt, add_parameters_from
from kqcircuits.util.refpoints import WaveguideToSimPort, JunctionSimPort


@add_parameters_from(Element, n=180)
class ConcentricTransmon(Qubit):
    """The PCell declaration for a concentric transmon.
    A concentric transmon consists of two islands, one inner and one outer, connected by a Josephson Junction/s.
    Multiple couplers can be defined. They can have custom waveguide impedance, size and shape.
    Each coupler has reference points, numbered starting from 1. Driveline can be connected to the drive port.

    """
    # Qubit geometry
    r_inner = Param(pdt.TypeDouble, "Internal island radius", 120, unit="μm",
                    docstring="Radius of the outer edge of the inner island")
    r_outer = Param(pdt.TypeDouble, "External island radius, measured at the outer edge", 250, unit="μm",
                    docstring="Radius of the external qubit island")
    outer_island_width = Param(pdt.TypeDouble, "Outer island radial width", 80, unit="μm",
                               docstring="Width of the external island")
    ground_gap = Param(pdt.TypeDouble, "Ground plane padding", 80, unit="μm")
    squid_angle = Param(pdt.TypeDouble, "Angular position of the Josephson Junction/s, where the positive x-axis",
                        120, unit="degrees")

    # Couplers parameters (the list size define the number of couplers)
    couplers_r = Param(pdt.TypeDouble, "Radius of the couplers positioning", 290, unit="[μm]")
    couplers_a = Param(pdt.TypeList, "Width of the coupler waveguide's center conductors", [10, 3, 4.5], unit="[μm]")
    couplers_b = Param(pdt.TypeList, "Width of the coupler waveguide's gaps", [6, 32, 20], unit="[μm]")
    couplers_angle = Param(pdt.TypeList,
                           "Positioning angles of the couplers, where 0deg corresponds to positive x-axis",
                           [340, 60, 210], unit="[degrees]")
    couplers_width = Param(pdt.TypeList, "Radial widths of the arc couplers", [10, 20, 30], unit="[μm]")
    couplers_arc_amplitude = Param(pdt.TypeList, "Couplers angular extension", [35, 45, 15], unit="[degrees]")

    # Drive port parameters
    drive_angle = Param(pdt.TypeDouble, "Angle of the drive port, where 0deg corresponds to positive x-axis", 300,
                        unit="degrees")
    drive_distance = Param(pdt.TypeDouble, "Distance of the driveline, measured from qubit centre", 400, unit="µm")

    def build(self):
        self.x_end = self.r_outer + self.ground_gap  # Define the outermost qubit coordinate

        # Generate the qubit islands (they are the negative shape of the final geometry)
        qubit_negative = self._make_qubit_islands()

        # Generate the coupler islands
        coupler_islands_region = self._make_coupler_island()

        # Add the waveguides connecting the couplers to external waveguides
        waveguide, waveguide_gap = self._make_waveguides()

        # Add the Josephson Junction/s
        self._add_junction(qubit_negative)

        # Define the capacitor in the ground (final polarity)
        ground_region = self._make_ground_region()
        qubit = ground_region - qubit_negative + waveguide_gap - coupler_islands_region - \
                waveguide  # Operations order is important!
        self.cell.shapes(self.get_layer("base_metal_gap_wo_grid")).insert(qubit)

        # Protection region from the ground grid
        region_protection = self._get_protection_region(ground_region)
        self.cell.shapes(self.get_layer("ground_grid_avoidance")).insert(region_protection)

        # Couplers and driveline ports
        self._add_ports()

    def _make_arc_island(self, island_outer_radius, island_width, swept_angle):
        # Generate a polygon arc of any size and angle
        angle_rad = math.radians(swept_angle)
        points_outside = arc_points(island_outer_radius, -angle_rad / 2, angle_rad / 2, self.n)
        points_inside = arc_points(island_outer_radius - island_width, angle_rad / 2, -angle_rad / 2, self.n)
        points = points_inside + points_outside
        arc_island = pya.DPolygon(points)
        return arc_island

    def _make_qubit_islands(self):
        # Generate a region of the qubit shunting capacitor
        capacitor_islands = []
        # Inner circular island
        inner_island = circle_polygon(self.r_inner, self.n)
        capacitor_islands.append(inner_island)
        # Outer cylindrical island
        outer_island = self._make_arc_island(self.r_outer, self.outer_island_width, 360)
        capacitor_islands.append(outer_island)
        return pya.Region([poly.to_itype(self.layout.dbu) for poly in capacitor_islands])

    def _make_coupler_island(self):
        # Generate the regions of the coupler islands.
        round_corner = 5
        coupler_islands_region = pya.Region()
        # Generate all the couplers in the same region
        for (c_angle, c_width, c_arc_ampl) in zip(self.couplers_angle, self.couplers_width,
                                                  self.couplers_arc_amplitude):
            coupler_island = self._make_arc_island(self.couplers_r + float(c_width) / 2, float(c_width),
                                                   float(c_arc_ampl))
            coupler_island_region = pya.Region(coupler_island.to_itype(self.layout.dbu)).round_corners(
                round_corner / self.layout.dbu, round_corner / self.layout.dbu, self.n).transformed(
                pya.ICplxTrans(1, float(c_angle), False, 0, 0))
            coupler_islands_region += coupler_island_region
        return coupler_islands_region

    def _make_waveguides(self):
        # Make the waveguides for each coupler with custom impedance and return the region
        waveguides_signal_region = pya.Region()
        waveguides_gap_region = pya.Region()
        # Add the waveguides inside the ground padding
        overlapping_margin = 0.5
        for (c_a, c_b, c_angle) in zip(self.couplers_a, self.couplers_b, self.couplers_angle):
            waveguide_signal = pya.Region([pya.DPolygon([
                pya.DPoint(self.x_end + overlapping_margin, float(c_a) / 2),
                pya.DPoint(self.couplers_r, float(c_a) / 2),
                pya.DPoint(self.couplers_r, -float(c_a) / 2),
                pya.DPoint(self.x_end + overlapping_margin, -float(c_a) / 2),
            ]).to_itype(self.layout.dbu)]).transformed(pya.ICplxTrans(1, float(c_angle), False, 0, 0))
            waveguide_gap = pya.Region([pya.DPolygon([
                pya.DPoint(self.x_end, float(c_a) / 2 + float(c_b)),
                pya.DPoint(self.couplers_r, float(c_a) / 2 + float(c_b)),
                pya.DPoint(self.couplers_r, -float(c_a) / 2 - float(c_b)),
                pya.DPoint(self.x_end, -float(c_a) / 2 - float(c_b)),
            ]).to_itype(self.layout.dbu)]).transformed(pya.ICplxTrans(1, float(c_angle), False, 0, 0))
            waveguides_signal_region += waveguide_signal
            waveguides_gap_region += waveguide_gap
        return waveguides_signal_region, waveguides_gap_region

    def _add_junction(self, region):
        # Add the junction to the qubit islands
        squid_origin = \
            arc_points(self.r_outer - self.outer_island_width, self.squid_angle * math.pi / 180, 2 * math.pi, self.n,
                       pya.DPoint(0, 0))[0]
        squid_transf = pya.DCplxTrans(1, self.squid_angle + 90, False, squid_origin)
        self.produce_squid(squid_transf)
        squid_distance_from_centre = self.refpoints['squid_port_common'].distance(self.refpoints['base'])
        # Connect the junction to the inner island
        squid_connection = pya.Region(squid_transf * pya.DPolygon([
            pya.DPoint(-4, 0),
            pya.DPoint(-4, -squid_distance_from_centre - 0.5),
            pya.DPoint(4, -squid_distance_from_centre - 0.5),
            pya.DPoint(4, 0)]
        ).to_itype(self.layout.dbu))
        region += squid_connection

    def _make_ground_region(self):
        # Generate the ground region as a filled circle of the maximum size
        return pya.Region(circle_polygon(self.x_end, self.n).to_itype(self.layout.dbu))

    def _add_ports(self):
        # Add couplers ports
        for i, c_angle in enumerate(map(float, self.couplers_angle)):
            coupler_origin = arc_points(self.x_end, c_angle * math.pi / 180, 2 * math.pi, self.n, pya.DPoint(0, 0))[0]
            coupler_transf = pya.DCplxTrans(1, 90 + c_angle, False, coupler_origin)
            self.add_port(f"coupler_{i + 1}", coupler_transf * pya.DPoint(0, 0),
                          direction=pya.DVector(coupler_transf * pya.DPoint(0, 0)))
        # Add driveline port
        drive_origin = arc_points(self.drive_distance, self.drive_angle * math.pi / 180, 2 * math.pi, self.n,
                                  pya.DPoint(0, 0))[0]
        drive_transf = pya.DCplxTrans(1, 90 + self.drive_angle, False, drive_origin)
        self.add_port("drive", drive_transf * pya.DPoint(0, 0), direction=pya.DVector(drive_transf * pya.DPoint(0, 0)))

    def _get_protection_region(self, region):
        # Region which we don't want to cover with the automatically generated ground grid
        protection_region = region.sized(self.margin / self.layout.dbu, self.margin / self.layout.dbu, 2)
        return protection_region

    @classmethod
    def get_sim_ports(cls, simulation):
        ports = [JunctionSimPort()]
        return ports + [WaveguideToSimPort(f"port_coupler_{i+1}", side="bottom",
                                           a=simulation.couplers_a[i],
                                           b=simulation.couplers_b[i]) for i in range(len(simulation.couplers_angle))]



# This code is part of KQCircuits
# Copyright (C) 2021 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).


from math import pi

from kqcircuits.chips.chip import Chip
from kqcircuits.elements.meander import Meander
from kqcircuits.qubits.swissmon import Swissmon
from kqcircuits.elements.waveguide_coplanar import WaveguideCoplanar
from kqcircuits.elements.waveguide_coplanar_splitter import WaveguideCoplanarSplitter, t_cross_parameters
from kqcircuits.pya_resolver import pya
from kqcircuits.util.coupler_lib import cap_params
from kqcircuits.util.parameters import Param, pdt, add_parameters_from
from kqcircuits.junctions.junction import Junction


def _get_num_meanders(meander_length, turn_radius, meander_min_width):
    """Get the required number of meanders to create a meander element with the given parameters."""

    return int((meander_length - turn_radius * (pi - 2)) / (meander_min_width + turn_radius * (pi - 2)))


@add_parameters_from(Junction, "junction_type")
class SingleXmons(Chip):
    """The PCell declaration for a SingleXmons chip.

    The SingleXmons chip has 6 qubits, which are coupled by readout resonators to the same feedline. The feedline
    crosses the center of the chip horizontally.  Half of the qubits are above the feedline and half are below it.
    For each qubit, there is a chargeline connected to a launcher, but no fluxline. There can optionally be four test
    resonators between the qubits.

    Attributes:
        launchers: A dictionary where the keys are names of the launchers and values are tuples whose first elements
            are positions of the launchers.

        qubits_refpoints: A tuple where each element contains the refpoints for one of the qubits. The qubits are
            ordered such that 0,1,2 are the upper qubits (from left to right), while 3,4,5 are the lower qubits (from
            left to right).

    """
    readout_res_lengths = Param(pdt.TypeList, "Readout resonator lengths (six resonators)",
                                [5000, 5100, 5200, 5300, 5400, 5500])
    use_test_resonators = Param(pdt.TypeBoolean, "Use test resonators", True)
    test_res_lengths = Param(pdt.TypeList, "Test resonator lengths (four resonators)", [5200, 5400, 5600, 5800])
    n_fingers = Param(pdt.TypeList, "Number of fingers for test resonator couplers", [4, 4, 2, 4])
    l_fingers = Param(pdt.TypeList, "Length of fingers for test resonator couplers", [23.1, 9.9, 14.1, 10, 21])
    type_coupler = Param(pdt.TypeList, "Coupler type for test resonator couplers",
                         ["interdigital", "interdigital", "interdigital", "gap"])

    def build(self):
        """Produces a SingleXmons PCell."""

        self.produce_junction_tests(self.junction_type)
        self.launchers = self.produce_launchers("SMA8")
        self.qubits_refpoints = self._produce_qubits()

        feedline_x_distance = 1200
        if self.use_test_resonators:
            self._produce_feedline_and_test_resonators(feedline_x_distance)
        else:
            self._produce_feedline(feedline_x_distance)

        self._produce_readout_resonators()
        self._produce_chargelines()

    def _produce_waveguide(self, path, term2=0, turn_radius=None):
        """Produces a coplanar waveguide that follows the given path.

        Args:
            path: a DPath object determining the waveguide path
            term2: term2 of the waveguide
            turn_radius: turn_radius of the waveguide

        Returns:
            length of the produced waveguide

        """
        if turn_radius is None:
            turn_radius = self.r
        waveguide = self.add_element(WaveguideCoplanar,
            path=pya.DPath(path, 1),
            r=turn_radius,
            term2=term2,
        )
        self.insert_cell(waveguide)
        return waveguide

    def _produce_qubit(self, qubit_cell, center_x, center_y, rotation, name=None):
        """Produces a qubit in a SingleXmons chip.

        Args:
            qubit_cell: PCell of the qubit.
            center_x: X-coordinate of the center of the qubit.
            center_y: Y-coordinate of the center of the qubit.
            rotation: An integer which defines the rotation of the qubit in units of 90 degrees.
            name: A string containing the name of this qubit. Used to set the "id" property of the qubit instance.

        Returns:
            refpoints of the qubit.

        """
        qubit_trans = pya.DTrans(rotation, False, center_x, center_y)
        _, refpoints_abs = self.insert_cell(qubit_cell, qubit_trans, name, rec_levels=None)
        return refpoints_abs

    def _produce_qubits(self):
        """Produces six Swissmon qubits in predefined positions in a SingleXmons chip.

        Three qubits are above the feedline and three are below it. The produced qubits are at equal distances
        from the feedline, and the distances between qubits in the x-direction are equal. The qubits are centered around
        the center of the chip.

        Returns:
            A tuple containing the refpoints of the qubits. Each element in the tuple contains the refpoints for a
            single qubit. The qubits are ordered such that 0,1,2 are the upper qubits (from left to right), while
            3,4,5 are the lower qubits (from left to right).

        """
        qubit = self.add_element(Swissmon,
            fluxline_type="none",
            arm_length=[146] * 4,
            arm_width=[24] * 4,
            gap_width=[24] * 4,
            island_r=2,
            cpl_length=[0, 140, 0],
            cpl_width=[60, 24, 60],
            cpl_gap=[110, 102, 110],
            cl_offset=[200, 200],
        )
        qubit_spacing_x = 1100  # shortest x-distance between qubit centers on different sides of the feedline
        qubit_spacing_y = 2600  # shortest y-distance between qubit centers on different sides of the feedline
        qubits_center_x = 5e3 + 400  # the x-coordinate around which qubits are centered
        # qubits above the feedline, from left to right
        y_a = 5e3 + qubit_spacing_y / 2
        qb0_refpoints = self._produce_qubit(qubit, qubits_center_x - qubit_spacing_x * (3 / 2), y_a, 2, "qb_0")
        qb1_refpoints = self._produce_qubit(qubit, qubits_center_x + qubit_spacing_x / 2, y_a, 2, "qb_1")
        qb2_refpoints = self._produce_qubit(qubit, qubits_center_x + qubit_spacing_x * (5 / 2) - 200, y_a, 2, "qb_2")
        # qubits below the feedline, from left to right
        y_b = 5e3 - qubit_spacing_y / 2
        qb3_refpoints = self._produce_qubit(qubit, qubits_center_x - qubit_spacing_x * (5 / 2), y_b, 0, "qb_3")
        qb4_refpoints = self._produce_qubit(qubit, qubits_center_x - qubit_spacing_x * (1 / 2), y_b, 0, "qb_4")
        qb5_refpoints = self._produce_qubit(qubit, qubits_center_x + qubit_spacing_x * (3 / 2), y_b, 0, "qb_5")
        return qb0_refpoints, qb1_refpoints, qb2_refpoints, qb3_refpoints, qb4_refpoints, qb5_refpoints

    def _produce_readout_resonator(self, total_length, coupling_length, pos_start, above_feedline):
        """Produces a readout resonator coupled to a qubit.

        The resonator consists of waveguide going vertically from the qubit towards the feedline, then going along
        the feedline (coupled to feedline) and finally going away from the feedline as a meander.

        Args:
            total_length: A float defining the total length of the resonator waveguide.
            coupling_length: A float defining the length of the part of the resonator coupled to the feedline.
            pos_start: A DPoint defining the start position of the resonator, should be at one of the qubit ports.
            above_feedline: A boolean value telling if the qubit is above the feedline or not.

        """
        # We define a factor depending on which side of the feedline the qubit is on. This lets us define all resonators
        # in the same way.
        if above_feedline:
            factor = 1
        else:
            factor = -1
        turn_radius = 50
        distance_to_feedline = 27
        feedline_coupling_y = 5e3 + factor * distance_to_feedline
        meander_start_x = pos_start.x - (coupling_length + 2 * turn_radius)
        meander_start = pya.DPoint(meander_start_x, 5e3 + factor * distance_to_feedline +
                                   factor * 2 * turn_radius)
        # non-meandering part of the resonator
        coupler_waveguide = self._produce_waveguide([
            pos_start,
            pya.DPoint(pos_start.x, feedline_coupling_y),
            pya.DPoint(meander_start_x, feedline_coupling_y),
            meander_start
        ], turn_radius=turn_radius)
        len_coupler = coupler_waveguide.length()
        # meandering part of the resonator
        meander_length = total_length - len_coupler
        w = 350
        num_meanders = _get_num_meanders(meander_length, turn_radius, w)
        self.insert_cell(Meander,
            start=meander_start,
            end=meander_start + pya.DPoint(0, 2 * factor * turn_radius * (num_meanders + 1)),
            length=meander_length,
            meanders=num_meanders,
            r=turn_radius,
        )

    def _produce_readout_resonators(self):
        """Produces readout resonators for all the qubits in a SingleXmons chip."""
        readout_res_lengths = [float(length) for length in self.readout_res_lengths]  # from strings to floats
        self._produce_readout_resonator(readout_res_lengths[0], 400, self.qubits_refpoints[0]["port_cplr1"], True)
        self._produce_readout_resonator(readout_res_lengths[1], 400, self.qubits_refpoints[1]["port_cplr1"], True)
        self._produce_readout_resonator(readout_res_lengths[2], 400, self.qubits_refpoints[2]["port_cplr1"], True)
        self._produce_readout_resonator(readout_res_lengths[3], 400, self.qubits_refpoints[3]["port_cplr1"], False)
        self._produce_readout_resonator(readout_res_lengths[4], 400, self.qubits_refpoints[4]["port_cplr1"], False)
        self._produce_readout_resonator(readout_res_lengths[5], 400, self.qubits_refpoints[5]["port_cplr1"], False)

    def _produce_chargeline(self, pos_launcher, pos_port_drive, y_distance):
        """Produces a chargeline from a launcher to a qubit.

        The chargeline is defined in such a way that it works well for the geometry of the a SingleXmons chip.

        Args:
            pos_launcher: A DPoint representing the position of the launcher.
            pos_port_drive: A DPoint representing the position of "port_drive" of the qubit.
            y_distance: A float defining the y-distance of the second point of the chargeline from the launcher.

        """
        points = [pos_launcher, pya.DPoint(pos_port_drive.x, pos_launcher.y + y_distance), pos_port_drive]
        # if y_distance!=0, we use four points to define the chargeline, otherwise three points
        if y_distance != 0:
            points = [points[0]] + [pya.DPoint(pos_launcher.x, pos_launcher.y + y_distance)] + points[1:3]
        self._produce_waveguide(points, term2=self.b)

    def _produce_chargelines(self):
        """Produces chargelines for all of the qubits in a SingleXmons chip."""
        self._produce_chargeline(self.launchers["NW"][0], self.qubits_refpoints[0]["port_drive"], -1300)
        self._produce_chargeline(self.launchers["NE"][0], self.qubits_refpoints[1]["port_drive"], -1300)
        self._produce_chargeline(self.launchers["EN"][0], self.qubits_refpoints[2]["port_drive"], 0)
        self._produce_chargeline(self.launchers["WS"][0], self.qubits_refpoints[3]["port_drive"], 0)
        self._produce_chargeline(self.launchers["SW"][0], self.qubits_refpoints[4]["port_drive"], 1300)
        self._produce_chargeline(self.launchers["SE"][0], self.qubits_refpoints[5]["port_drive"], 1300)

    def _produce_test_resonator(self, capacitor, capacitor_dtrans, res_idx):

        factor = (2*(res_idx % 2) - 1)  # -1 for resonators below feedline, +1 for resonators above feedline
        total_length = float(self.test_res_lengths[res_idx])
        turn_radius = 50

        # non-meandering part of the resonator
        pos_start = self.get_refpoints(capacitor, capacitor_dtrans)["port_a"]
        x1 = 500
        y1 = factor * 300
        y2 = factor * 100
        meander_start = pos_start + pya.DPoint(x1, y1 + y2)
        nonmeander_waveguide = self._produce_waveguide([
            pos_start,
            pos_start + pya.DPoint(0, y1),
            pos_start + pya.DPoint(x1, y1),
            meander_start,
        ], turn_radius=turn_radius)
        len_nonmeander = nonmeander_waveguide.length()

        # meandering part of the resonator
        meander_length = total_length - len_nonmeander
        w = 250
        num_meanders = _get_num_meanders(meander_length, turn_radius, w)
        self.insert_cell(Meander,
            start=meander_start,
            end=meander_start + pya.DPoint(0, 2 * factor * turn_radius * (num_meanders + 1)),
            length=meander_length,
            meanders=num_meanders,
            r=turn_radius,
        )

    def _produce_feedline(self, x_distance):
        """Produces a feedline for a SingleXmons chip.

        The feedline is a waveguide connecting launcher "WN" to launcher "ES". It goes horizontally from the
        launchers towards the center of the chip until it is clear of the junction test pads. Then it goes vertically
        to the center horizontal line, after which it follows the horizontal line until the two sides are connected.

        Args:
            x_distance: A float defining the x-distance of the vertical parts from the launchers. This is used to stay
                clear from the junction test pads.

        """
        self._produce_waveguide([
            self.launchers["WN"][0],
            pya.DPoint(self.launchers["WN"][0].x + x_distance, self.launchers["WN"][0].y),
            pya.DPoint(self.launchers["WN"][0].x + x_distance, 5e3),
            pya.DPoint(self.launchers["ES"][0].x - x_distance, 5e3),
            pya.DPoint(self.launchers["ES"][0].x - x_distance, self.launchers["ES"][0].y),
            self.launchers["ES"][0]
        ])

    def _produce_feedline_and_test_resonators(self, x_distance):
        """Produces a feedline and test resonators for a SingleXmons chip.

        The feedline is a waveguide connecting launcher "WN" to launcher "ES". It goes horizontally from the
        launchers towards the center of the chip until it is clear of the junction test pads. Then it goes vertically
        to the center horizontal line, after which it follows the horizontal line until the two sides are connected.
        There are four test resonators, located between the qubit pairs.

        Args:
            x_distance: A float defining the x-distance of the vertical parts from the launchers. This is used to stay
                clear from the junction test pads.

        """
        x_offset = -700
        test_resonator_positions = [
            pya.DPoint((self.qubits_refpoints[3]["base"].x + self.qubits_refpoints[4]["base"].x) / 2 + x_offset, 5e3),
            pya.DPoint((self.qubits_refpoints[1]["base"].x + self.qubits_refpoints[0]["base"].x) / 2 + x_offset, 5e3),
            pya.DPoint((self.qubits_refpoints[5]["base"].x + self.qubits_refpoints[4]["base"].x) / 2 + x_offset, 5e3),
            pya.DPoint((self.qubits_refpoints[2]["base"].x + self.qubits_refpoints[1]["base"].x) / 2 + x_offset, 5e3)
        ]

        # feedline couplings with test resonators

        cell_cross = self.add_element(WaveguideCoplanarSplitter, **t_cross_parameters(
            a=self.a, b=self.b, a2=self.a, b2=self.b, length_extra_side=2 * self.a))
        inst_crosses = []

        for i in range(4):
            # Cross
            cross_trans = pya.DTrans(2 * (i % 2), False, test_resonator_positions[i])
            inst_cross, _ = self.insert_cell(cell_cross, cross_trans)
            inst_crosses.append(inst_cross)
            cross_refpoints_abs = self.get_refpoints(cell_cross, inst_crosses[i].dtrans)

            # Coupler
            cplr_params = cap_params(float(self.n_fingers[i]), float(self.l_fingers[i]), self.type_coupler[i])
            cplr = self.add_element(**cplr_params)
            cplr_refpoints_rel = self.get_refpoints(cplr)
            if i % 2 == 0:
                cplr_pos = cross_refpoints_abs["port_bottom"] - pya.DTrans.R90 * cplr_refpoints_rel["port_b"]
            else:
                cplr_pos = cross_refpoints_abs["port_bottom"] + pya.DTrans.R90 * cplr_refpoints_rel["port_b"]
            cplr_dtrans = pya.DTrans(2 * (i % 2) + 1, False, cplr_pos.x, cplr_pos.y)
            self.insert_cell(cplr, cplr_dtrans)

            self._produce_test_resonator(cplr, cplr_dtrans, i)

        # feedline

        self._produce_waveguide([
            self.launchers["WN"][0],
            pya.DPoint(self.launchers["WN"][0].x + x_distance, self.launchers["WN"][0].y),
            pya.DPoint(self.launchers["WN"][0].x + x_distance, 5e3),
            self.get_refpoints(cell_cross, inst_crosses[0].dtrans)["port_left"],
            ])
        self._produce_waveguide([
            self.get_refpoints(cell_cross, inst_crosses[0].dtrans)["port_right"],
            self.get_refpoints(cell_cross, inst_crosses[1].dtrans)["port_right"],
           ])
        self._produce_waveguide([
            self.get_refpoints(cell_cross, inst_crosses[1].dtrans)["port_left"],
            self.get_refpoints(cell_cross, inst_crosses[2].dtrans)["port_left"],
            ])
        self._produce_waveguide([
            self.get_refpoints(cell_cross, inst_crosses[2].dtrans)["port_right"],
            self.get_refpoints(cell_cross, inst_crosses[3].dtrans)["port_right"],
        ])
        self._produce_waveguide([
            self.get_refpoints(cell_cross, inst_crosses[3].dtrans)["port_left"],
            pya.DPoint(self.launchers["ES"][0].x - x_distance, 5e3),
            pya.DPoint(self.launchers["ES"][0].x - x_distance, self.launchers["ES"][0].y),
            self.launchers["ES"][0]
        ])



# This code is part of KQCircuits
# Copyright (C) 2022 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).


from kqcircuits.util.geometry_helper import bspline_points
from kqcircuits.util.parameters import Param, pdt
from kqcircuits.qubits.double_pads import DoublePads
from kqcircuits.pya_resolver import pya

class DoublePadsSplines(DoublePads):
    """A two-island qubit, consisting of two spline-constructed islands shunted by a junction,
    with one capacitive coupler.

    Contains a coupler on the north edge and two separate qubit islands in the center
    joined by a junction or SQUID loaded from another library.
    Refpoint for a readout line at the opening to the coupler and a modifiable refpoint for
    a driveline outside of the rectangle.
    """

    island_spline = Param(pdt.TypeList, "Control points of a B-Spline that defines left half of the island",
        [-250, 20, -250, 80, -100, 100, 0, 100])
    island_spline_samples = Param(pdt.TypeInt, "Number of samples taken from each island spline", 100,
        docstring=( "Number of samples taken from each island spline. "
                    "There is a spline for each consequent series of four control points"))


    def _build_island1(self, squid_height, taper_height):
        island1_bottom = self.squid_offset + squid_height / 2
        curve_start = pya.DPoint(0, island1_bottom + taper_height)
        island1_control_points = [curve_start + pya.DPoint(-self.island1_taper_width / 2, 0)] + \
            [curve_start + pya.DPoint(float(self.island_spline[idx]), float(self.island_spline[idx+1]))
                for idx in range(0, len(self.island_spline), 2)]
        island1_control_points += [pya.DPoint(-p.x, p.y) for p in reversed(island1_control_points[:-1])]
        for i, p in enumerate(island1_control_points):
            self.refpoints[f"island1_control_point_{i}"] = p
        island1_polygon = pya.DPolygon(bspline_points(island1_control_points,
                                                        self.island_spline_samples,
                                                        startpoint=True, endpoint=True))
        island1_region = pya.Region(island1_polygon.to_itype(self.layout.dbu))
        island1_taper = pya.Region(pya.DPolygon([
            #pertrude taper into island in case of precision errors
            pya.DPoint( self.island1_taper_width / 2, island1_bottom + taper_height + 1),
            pya.DPoint( self.island1_taper_width / 2, island1_bottom + taper_height),
            pya.DPoint( self.island1_taper_junction_width / 2, island1_bottom),
            pya.DPoint(-self.island1_taper_junction_width / 2, island1_bottom),
            pya.DPoint(-self.island1_taper_width / 2, island1_bottom + taper_height),
            #pertrude taper into island in case of precision errors
            pya.DPoint(-self.island1_taper_width / 2, island1_bottom + taper_height + 1),
        ]).to_itype(self.layout.dbu))
        return island1_region + island1_taper


    def _build_island2(self, squid_height, taper_height):
        island2_top = self.squid_offset - squid_height / 2
        curve_start = pya.DPoint(0, island2_top - taper_height)
        island2_control_points = [curve_start + pya.DPoint(-self.island2_taper_width / 2, 0)] + \
            [curve_start + pya.DPoint(float(self.island_spline[idx]), -float(self.island_spline[idx+1]))
                for idx in range(0, len(self.island_spline), 2)]
        island2_control_points += [pya.DPoint(-p.x, p.y) for p in reversed(island2_control_points[:-1])]
        for i, p in enumerate(island2_control_points):
            self.refpoints[f"island2_control_point_{i}"] = p
        island2_polygon = pya.DPolygon(bspline_points(island2_control_points,
                                                        self.island_spline_samples,
                                                        startpoint=True, endpoint=True))
        island2_region = pya.Region(island2_polygon.to_itype(self.layout.dbu))
        island2_taper = pya.Region(pya.DPolygon([
            #pertrude taper into island in case of precision errors
            pya.DPoint( self.island2_taper_width / 2, island2_top - taper_height - 1),
            pya.DPoint( self.island2_taper_width / 2, island2_top - taper_height),
            pya.DPoint( self.island2_taper_junction_width / 2, island2_top),
            pya.DPoint(-self.island2_taper_junction_width / 2, island2_top),
            pya.DPoint(-self.island2_taper_width / 2, island2_top - taper_height),
            #pertrude taper into island in case of precision errors
            pya.DPoint(-self.island2_taper_width / 2, island2_top - taper_height - 1),
        ]).to_itype(self.layout.dbu))
        return island2_region + island2_taper



# This code is part of KQCircuits
# Copyright (C) 2021 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).


from kqcircuits.elements.element import Element
from kqcircuits.pya_resolver import pya


class TestStructure(Element):
    """Base PCell declaration for test structures."""

    LIBRARY_NAME = "Test Structure Library"
    LIBRARY_DESCRIPTION = "Superconducting quantum circuit library for test structures."
    LIBRARY_PATH = "test_structures"

    def produce_pad(self, x, y, pads_region, pad_width, pad_height):
        """Inserts a square pad shape to pads_region.

        Args:
            x: x-coordinate of the pad center
            y: y-coordinate of the pad center
            pads_region: Region to which the pad shape is inserted
            pad_width: width (and height) of the pad

        """
        offset_x = pad_width/2
        offset_y = pad_height/2
        pad = pya.DPolygon([
            pya.DPoint(x - offset_x, y - offset_y),
            pya.DPoint(x - offset_x, y + offset_y),
            pya.DPoint(x + offset_x, y + offset_y),
            pya.DPoint(x + offset_x, y - offset_y),
        ])
        pads_region.insert(pad.to_itype(self.layout.dbu))

    def produce_four_point_pads(self, pads_region, pad_width, pad_height, pad_spacing_x, pad_spacing_y, connect_pads,
                                trans=pya.DTrans(), refpoint_prefix="probe", refpoint_distance=None):
        """Inserts four pads to pads_region.

        Args:
            pads_region: Region to which the pad shapes are inserted
            pad_width: width (and height) of the pads
            pad_spacing_x: x-distance between the pads
            pad_spacing_y: y-distance between the pads
            connect_pads: Boolean determining if the two pads on each side are connected
            trans: transformation applied to the pads
            refpoint_prefix: prefix used for the refpoint names
            refpoint_distance: Horizontal distance of refpoints from closest outer edges.
                If None, the refpoints will be at the center of the pads.

        """

        pad_offset_x = (pad_spacing_x + pad_width)/2
        pad_offset_y = (pad_spacing_y + pad_height)/2

        pos_sw = pya.DPoint(trans.disp.x - pad_offset_x, trans.disp.y - pad_offset_y)
        pos_nw = pya.DPoint(trans.disp.x - pad_offset_x, trans.disp.y + pad_offset_y)
        pos_ne = pya.DPoint(trans.disp.x + pad_offset_x, trans.disp.y + pad_offset_y)
        pos_se = pya.DPoint(trans.disp.x + pad_offset_x, trans.disp.y - pad_offset_y)

        self.produce_pad(pos_sw.x, pos_sw.y, pads_region, pad_width, pad_height)
        self.produce_pad(pos_nw.x, pos_nw.y, pads_region, pad_width, pad_height)
        self.produce_pad(pos_ne.x, pos_ne.y, pads_region, pad_width, pad_height)
        self.produce_pad(pos_se.x, pos_se.y, pads_region, pad_width, pad_height)

        if connect_pads:
            pad_connection_box_width = 50
            pad_connection_box = pya.DBox(pya.DPoint(0, 0),
                                          pya.DPoint(pad_connection_box_width, pad_spacing_y))
            trans_left = pya.DTrans(-pad_spacing_x/2 - pad_connection_box_width, -pad_spacing_y/2)
            trans_right = pya.DTrans(pad_spacing_x/2, -pad_spacing_y/2)
            pads_region.insert((trans*trans_left*pad_connection_box).to_itype(self.layout.dbu))
            pads_region.insert((trans*trans_right*pad_connection_box).to_itype(self.layout.dbu))

        if refpoint_distance is None:
            self.refpoints["{}_sw".format(refpoint_prefix)] = pos_sw
            self.refpoints["{}_nw".format(refpoint_prefix)] = pos_nw
            self.refpoints["{}_ne".format(refpoint_prefix)] = pos_ne
            self.refpoints["{}_se".format(refpoint_prefix)] = pos_se
        else:
            self.refpoints["{}_sw".format(refpoint_prefix)] = pos_sw + pya.DVector(-pad_width/2 + refpoint_distance, 0)
            self.refpoints["{}_nw".format(refpoint_prefix)] = pos_nw + pya.DVector(-pad_width/2 + refpoint_distance, 0)
            self.refpoints["{}_ne".format(refpoint_prefix)] = pos_ne + pya.DVector(pad_width/2 - refpoint_distance, 0)
            self.refpoints["{}_se".format(refpoint_prefix)] = pos_se + pya.DVector(pad_width/2 - refpoint_distance, 0)

    def produce_etched_region(self, metal_region, pos, width, height):
        """Produces structures in metal gap layer.

        Subtracts metal_region from a region defined by width and height, and inserts resulting
        region to metal gap layer. Also adds grid avoidance for the area.

        Args:
            metal_region: region subtracted from the region in metal gap layer
            pos: center of the produced region (pya.DPoint)
            width: width of the produced region
            height: height of the produced region

        """
        test_area = pya.DPolygon([
            pya.DPoint(pos.x + width/2, pos.y + height/2),
            pya.DPoint(pos.x + width/2, pos.y - height/2),
            pya.DPoint(pos.x - width/2, pos.y - height/2),
            pya.DPoint(pos.x - width/2, pos.y + height/2),
        ])
        reg_test_area = pya.Region(test_area.to_itype(self.layout.dbu))
        # etched region
        reg_etch = reg_test_area - metal_region
        self.cell.shapes(self.get_layer("base_metal_gap_wo_grid")).insert(reg_etch)
        # grid avoidance region
        reg_protect = reg_etch.extents(int(self.margin/self.layout.dbu))
        self.cell.shapes(self.get_layer("ground_grid_avoidance")).insert(reg_protect)



# This code is part of KQCircuits
# Copyright (C) 2021 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).
from kqcircuits.elements.element import Element
from kqcircuits.pya_resolver import pya
from kqcircuits.util.geometry_helper import region_with_merged_polygons


def merge_layers(layout, cell_list, layer_1, layer_2, layer_merged):
    """Inserts shapes in layer_1 and layer_2 to layer_merged.

    Args:
        layout: Layout of the cells
        cell_list: list of Cells whose shapes are merged (recursively)
        layer_1: LayerInfo of the layer to be merged with layer_2
        layer_2: LayerInfo of the layer to be merged with layer_1
        layer_merged: LayerInfo of the layer where the merged shapes are inserted

    """
    for cell in cell_list:
        iter1 = pya.RecursiveShapeIterator(layout, cell, layout.layer(layer_1))
        iter2 = pya.RecursiveShapeIterator(layout, cell, layout.layer(layer_2))

        reg1 = pya.Region(iter1)
        reg2 = pya.Region(iter2)

        merge_reg = reg1 + reg2

        cell.shapes(layout.layer(layer_merged)).insert(merge_reg)


def merge_layout_layers_on_face(layout, cell, face, tolerance=0.004):
    """Creates "base_metal_gap" layer on given face.

    The layer shape is combination of three layers using subtract (-) and insert (+) operations:

        "base_metal_gap" = "base_metal_gap_wo_grid" - "base_metal_addition" + "ground_grid"

    Args:
        layout: Layout containing the cell
        cell: Cell to merge
        face: face dictionary containing layer names as keys and layer info objects as values
        tolerance: gap length to be ignored while merging (µm)
    """
    gaps = pya.Region(cell.begin_shapes_rec(layout.layer(face["base_metal_gap_wo_grid"])))
    metal = pya.Region(cell.begin_shapes_rec(layout.layer(face["base_metal_addition"])))
    grid = cell.begin_shapes_rec(layout.layer(face["ground_grid"]))
    res = cell.shapes(layout.layer(face["base_metal_gap"]))
    res.insert(region_with_merged_polygons(gaps - metal, tolerance / layout.dbu))
    res.insert(grid)


def convert_child_instances_to_static(layout: pya.Layout, cell: pya.Cell, only_elements: bool = True,
                                      prune: bool = True):
    """Convert child instances of a cell to static.

    This function avoids duplicating cells: in case there are multiple instances pointing to the same PCell, only
    one static cell is created and all instances are pointed to that cell. This is in contrast to calling
    ``Instance.convert_to_static()`` on each instance separately.

    Args:
        layout: the Layout that contains the cell
        cell: static cell that may contain PCell instances to be converted
        only_elements: if True (default), only PCells which are descendants of Element are made static
        prune: if True, the PCells are deleted if they are no longer used anywhere in the layout.
    """

    # Build a dictionary {pcell: list of instances} for all instances that need to be converted
    cells_to_convert = {}
    for inst in cell.each_inst():
        if inst.is_pcell() and (not only_elements or isinstance(inst.pcell_declaration(), Element)):
            pcell = layout.cell(inst.cell_index)
            cells_to_convert[pcell] = cells_to_convert.get(pcell, []) + [inst]

    # Convert the cells and point the instances to the new static cell
    for pcell, instances in cells_to_convert.items():
        static_cell_index = layout.convert_cell_to_static(pcell.cell_index())
        for inst in instances:
            inst.cell_index = static_cell_index

    # Prune: deletes the old PCells unless they are still used elsewhere.
    if prune:
        for pcell in cells_to_convert:
            layout.prune_cell(pcell.cell_index(), -1)



# This code is part of KQCircuits
# Copyright (C) 2022 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).


import os
import stat
import logging
import json
import argparse

from pathlib import Path
from distutils.dir_util import copy_tree
from typing import Sequence

from kqcircuits.simulations.export.util import export_layers
from kqcircuits.util.export_helper import write_commit_reference_file
from kqcircuits.defaults import ELMER_SCRIPT_PATHS, KQC_REMOTE_ACCOUNT
from kqcircuits.simulations.simulation import Simulation
from kqcircuits.simulations.cross_section_simulation import CrossSectionSimulation
from kqcircuits.util.geometry_json_encoder import GeometryJsonEncoder


def copy_elmer_scripts_to_directory(path: Path):
    """
    Copies Elmer scripts into directory path.

    Args:
        path: Location where to copy scripts folder.
    """
    if path.exists() and path.is_dir():
        for script_path in ELMER_SCRIPT_PATHS:
            copy_tree(str(script_path), str(path), update=1)


def export_elmer_json(simulation,
                      path: Path,
                      tool='capacitance',
                      linear_system_method='bicgstab',
                      p_element_order=1,
                      frequency=5,
                      mesh_size=None,
                      boundary_conditions=None,
                      workflow=None,
                      percent_error=0.005,
                      max_error_scale=2,
                      max_outlier_fraction=1e-3,
                      maximum_passes=1,
                      minimum_passes=1,
                      dielectric_surfaces=None,
                      is_axisymmetric=False):
    """
    Export Elmer simulation into json and gds files.

    Args:
        simulation: The simulation to be exported.
        path: Location where to write json.
        tool(str): Available: "capacitance", "wave_equation" and "cross-section" (Default: capacitance)
        linear_system_method(str): Available: 'bicgstab', 'mg' (Default: bicgstab)
        p_element_order(int): polynomial order of p-elements (Default: 1)
        frequency: Units are in GHz. To set up multifrequency analysis, use list of numbers.
        mesh_size(dict): Parameters to determine mesh element sizes
        boundary_conditions(dict): Parameters to determine boundary conditions
        workflow(dict): Parameters for simulation workflow
        percent_error(float): Stopping criterion in adaptive meshing.
        max_error_scale(float): Maximum element error, relative to percent_error, allowed in individual elements.
        max_outlier_fraction(float): Maximum fraction of outliers from the total number of elements
        maximum_passes(int): Maximum number of adaptive meshing iterations.
        minimum_passes(int): Minimum number of adaptive meshing iterations.
        dielectric_surfaces: Loss tangents for dielectric interfaces, thickness and permittivity should be specified in
            the simulation. The loss tangent is post-processed to the participation to get the quality factor.
            Default is None. Input is of the form::

                'substrate': {
                    'tan_delta_surf': 5e-7
                },
                'layerMA': {  # metal–vacuum
                    'tan_delta_surf': 0.001,  # loss tangent
                },
                'layerMS': { # metal–substrate
                    'tan_delta_surf': 0.001,
                },
                'layerSA': { # substrate–vacuum
                    'tan_delta_surf': 0.001,
                }
        is_axisymmetric(bool): Simulate with Axi Symmetric coordinates along :math:`y\\Big|_{x=0}` (Default: False)

    Returns:
         Path to exported json file.
    """
    is_cross_section = isinstance(simulation, CrossSectionSimulation)

    if simulation is None or not isinstance(simulation, (Simulation, CrossSectionSimulation)):
        raise ValueError("Cannot export without simulation")

    # collect data for .json file
    if is_cross_section:
        layers = simulation.layer_dict

    if not isinstance(frequency, list):
        frequency = [frequency]

    json_data = {
        'tool': tool,
        **simulation.get_simulation_data(),
        **({'layers': {k: (v.layer, v.datatype) for k, v in layers.items()}} if is_cross_section else {}),
        'mesh_size': {} if mesh_size is None else mesh_size,
        'boundary conditions': boundary_conditions,
        'workflow': {} if workflow is None else workflow,
        'percent_error': percent_error,
        'max_error_scale': max_error_scale,
        'max_outlier_fraction': max_outlier_fraction,
        'maximum_passes': maximum_passes,
        'minimum_passes': minimum_passes,
        'frequency': frequency,
        **({} if dielectric_surfaces is None else {'dielectric_surfaces': dielectric_surfaces}),
        'linear_system_method': linear_system_method,
        'p_element_order': p_element_order,
        'is_axisymmetric': is_axisymmetric,
    }

    if is_cross_section:
        if json_data.get('london_penetration_depth', 0.0) > 0:
            sif_names = ['capacitance', 'inductance']
        else:
            sif_names = ['capacitance', 'capacitance0']
    elif tool=='wave_equation':
        stem = json_data['parameters']['name']
        sif_names = [stem + '_f' + str(f).replace('.', '_') for f in frequency]
    else:
        sif_names = [json_data['parameters']['name']]

    json_data['sif_names'] = sif_names

    # write .json file
    json_filename = str(path.joinpath(simulation.name + '.json'))
    with open(json_filename, 'w') as fp:
        json.dump(json_data, fp, cls=GeometryJsonEncoder, indent=4)

    # write .gds file
    gds_filename = str(path.joinpath(simulation.name + '.gds'))
    export_layers(
        gds_filename,
        simulation.layout,
        [simulation.cell],
        output_format='GDS2',
        layers=layers.values() if is_cross_section else simulation.get_layers()
    )

    return json_filename


def export_elmer_script(json_filenames, path: Path, workflow=None, file_prefix='simulation',
        script_file='scripts/run.py'):
    """
    Create script files for running one or more simulations.
    Create also a main script to launch all the simulations at once.

    Args:
        json_filenames: List of paths to json files to be included into the script.
        path: Location where to write the script file.
        workflow(dict): Parameters for simulation workflow
        file_prefix: Name of the script file to be created.
        script_file: Name of the script file to run.

    Returns:

        Path of exported main script file
    """
    if workflow is None:
        workflow = dict()
    sbatch = 'sbatch_parameters' in workflow
    python_executable = workflow.get('python_executable', 'python')
    parallelize_workload = 'n_workers' in workflow

    main_script_filename = str(path.joinpath(file_prefix + '.sh'))

    if sbatch:
        sbatch_parameters = workflow['sbatch_parameters']

        if sbatch_parameters.get('--account', 'project_0') == 'project_0':
            sbatch_parameters['--account'] = KQC_REMOTE_ACCOUNT

        common_keys = [k for k in sbatch_parameters.keys() if k.startswith('--')]
        sbatch_settings_elmer = {k: sbatch_parameters.pop(k) for k in common_keys}

        sbatch_settings_meshes = sbatch_settings_elmer.copy()

        sbatch_settings_elmer['--time'] = sbatch_parameters.pop('elmer_time', '00:10:00')
        sbatch_settings_elmer['--nodes'] = sbatch_parameters.pop('elmer_n_nodes', 1)
        sbatch_settings_elmer['--ntasks'] = sbatch_parameters.pop('elmer_n_processes', 10)
        sbatch_settings_elmer['--cpus-per-task'] = sbatch_parameters.pop('elmer_n_threads', 1)
        sbatch_settings_elmer['--mem'] = sbatch_parameters.pop('elmer_mem', '64G')

        sbatch_settings_meshes['--time'] = sbatch_parameters.pop('gmsh_time', '00:10:00')
        sbatch_settings_meshes['--nodes'] = 1
        sbatch_settings_meshes['--ntasks'] = 1
        sbatch_settings_meshes['--cpus-per-task'] = sbatch_parameters.pop('gmsh_n_threads', 10)
        sbatch_settings_meshes['--mem'] = sbatch_parameters.pop('gmsh_mem', '64G')

        if len(sbatch_parameters) > 0:
            logging.warning("Unused sbatch parameters: ")
            for k, v in sbatch_parameters.items():
                logging.warning(f"{k} : {v}")

        gmsh_n_threads = sbatch_settings_meshes['--cpus-per-task']
        elmer_n_processes = sbatch_settings_elmer['--ntasks']
        elmer_n_threads = sbatch_settings_elmer['--cpus-per-task']

        main_script_filename_meshes = str(path.joinpath(file_prefix + '_meshes.sh'))
        with open(main_script_filename_meshes, 'w') as main_file:
            main_file.write('#!/bin/bash\n')

            for s_key, s_value in sbatch_settings_meshes.items():
                main_file.write(f'#SBATCH {s_key}={s_value}\n')

            main_file.write('\n')
            main_file.write('# set the number of threads based on --cpus-per-task\n')
            main_file.write('export OMP_NUM_THREADS=$SLURM_CPUS_PER_TASK\n')
            main_file.write('\n')

            for i, json_filename in enumerate(json_filenames):
                with open(json_filename) as f:
                    json_data = json.load(f)
                    simulation_name = json_data['parameters']['name']

                sif_names = json_data['sif_names']

                script_filename_meshes = str(path.joinpath(simulation_name + '_meshes.sh'))
                with open(script_filename_meshes, 'w') as file:
                    file.write('echo "Simulation {}/{} Gmsh"\n'.format(i + 1, len(json_filenames)))
                    file.write('srun -c {3} {2} "{0}" "{1}" --only-gmsh -q 2>&1 >> '\
                            '"{1}_Gmsh.log"\n'.format(
                        script_file,
                        Path(json_filename).relative_to(path),
                        python_executable,
                        str(gmsh_n_threads)
                        ))
                    file.write('echo "Simulation {}/{} ElmerGrid"\n'.format(i + 1, len(json_filenames)))
                    file.write('srun -c {2} ElmerGrid 14 2 "{0}" 2>&1 >> "{1}_ElmerGrid.log"\n'.format(
                        simulation_name + ".msh",
                        Path(json_filename).relative_to(path),
                        str(gmsh_n_threads)
                    ))
                    if int(elmer_n_processes) > 1:
                        file.write('srun -c {3} ElmerGrid 2 2 "{0}" 2>&1 -metis {1} 4 --partdual --removeunused >> '\
                                '"{2}_ElmerGrid_part.log"\n'.format(
                            simulation_name,
                            str(elmer_n_processes),
                            Path(json_filename).relative_to(path),
                            str(gmsh_n_threads)
                        ))
                    file.write('echo "Simulation {}/{} Write Elmer sif files"\n'.format(i + 1, len(json_filenames)))
                    file.write('srun -c {3} {2} "{0}" "{1}" --only-elmer-sifs -q 2>&1 >> "{1}_Elmer.log"\n'.format(
                        script_file,
                        Path(json_filename).relative_to(path),
                        python_executable,
                        str(gmsh_n_threads)
                    ))
                os.chmod(script_filename_meshes, os.stat(script_filename_meshes).st_mode | stat.S_IEXEC)

                main_file.write('echo "Submitting gmsh and ElmerGrid part {}/{}"\n'
                            .format(i + 1, len(json_filenames)))
                main_file.write('echo "--------------------------------------------"\n')
                main_file.write('source "{}"\n'.format(Path(script_filename_meshes).relative_to(path)))

        # change permission
        os.chmod(main_script_filename_meshes, os.stat(main_script_filename_meshes).st_mode | stat.S_IEXEC)

        with open(main_script_filename, 'w') as main_file:
            main_file.write('#!/bin/bash\n')

            for s_key, s_value in sbatch_settings_elmer.items():
                main_file.write(f'#SBATCH {s_key}={s_value}\n')

            main_file.write('\n')
            main_file.write('# set the number of threads based on --cpus-per-task\n')
            main_file.write('export OMP_NUM_THREADS=$SLURM_CPUS_PER_TASK\n')
            main_file.write('\n')
            for i, json_filename in enumerate(json_filenames):
                with open(json_filename) as f:
                    json_data = json.load(f)
                    simulation_name = json_data['parameters']['name']

                sif_names = json_data['sif_names']
                script_filename = str(path.joinpath(simulation_name + '.sh'))
                with open(script_filename, 'w') as file:

                    file.write('echo "Simulation {}/{} Elmer"\n'.format(i + 1, len(json_filenames)))
                    for sif in sif_names:
                        file.write('srun --cpu-bind none -n {0} -c {4} ElmerSolver_mpi '\
                                '"{1}/{2}.sif" 2>&1 >> "{3}_Elmer.log"\n'.format(
                                    elmer_n_processes,
                                    simulation_name,
                                    sif,
                                    Path(json_filename).relative_to(path),
                                    elmer_n_threads
                                    ))
                    file.write('echo "Simulation {}/{} write results json"\n'.format(i + 1, len(json_filenames)))
                    file.write('srun -n 1 -c {3} {2} "{0}" "{1}" --write-project-results 2>&1 >> '\
                            '{1}_write_project_results.log\n'.format(
                        script_file,
                        Path(json_filename).relative_to(path),
                        python_executable,
                        elmer_n_threads
                    ))

                    file.write('echo "Simulation {}/{} write versions json"\n'.format(i + 1, len(json_filenames)))
                    file.write('srun -n 1 -c {3} {2} "{0}" "{1}" --write-versions-file 2>&1 >> '\
                            '{1}_write_versions_file.log\n'.format(
                        script_file,
                        Path(json_filename).relative_to(path),
                        python_executable,
                        elmer_n_threads
                    ))
                os.chmod(script_filename, os.stat(script_filename).st_mode | stat.S_IEXEC)

            main_file.write('echo "Submitting ElmerSolver part{}/{}"\n'
                        .format(i + 1, len(json_filenames)))
            main_file.write('echo "--------------------------------------------"\n')
            main_file.write('source "{}"\n'.format(Path(script_filename).relative_to(path)))


    else:

        with open(main_script_filename, 'w') as main_file:

            if parallelize_workload and not sbatch:
                main_file.write('{} scripts/simple_workload_manager.py {}'.format(
                    python_executable,
                    workflow['n_workers']
                ))

            for i, json_filename in enumerate(json_filenames):
                with open(json_filename) as f:
                    json_data = json.load(f)
                    simulation_name = json_data['parameters']['name']

                sif_names = json_data['sif_names']

                script_filename = str(path.joinpath(simulation_name + '.sh'))
                with open(script_filename, 'w') as file:
                    file.write('echo "Simulation {}/{} Gmsh"\n'.format(i + 1, len(json_filenames)))
                    file.write('{2} "{0}" "{1}" --only-gmsh 2>&1 >> "{1}_Gmsh.log"\n'.format(
                        script_file,
                        Path(json_filename).relative_to(path),
                        python_executable)
                    )
                    file.write('echo "Simulation {}/{} ElmerGrid"\n'.format(i + 1, len(json_filenames)))
                    file.write('{2} "{0}" "{1}" --only-elmergrid 2>&1 >> "{1}_ElmerGrid.log"\n'.format(
                        script_file,
                        Path(json_filename).relative_to(path),
                        python_executable)
                    )
                    file.write('echo "Simulation {}/{} Write Elmer sif files"\n'.format(i + 1, len(json_filenames)))
                    file.write('{2} "{0}" "{1}" --only-elmer-sifs 2>&1 >> "{1}_Elmer.log"\n'.format(
                        script_file,
                        Path(json_filename).relative_to(path),
                        python_executable)
                        )
                    file.write('echo "Simulation {}/{} Elmer"\n'.format(i + 1, len(json_filenames)))
                    file.write('{2} "{0}" "{1}" --only-elmer 2>&1 >> "{1}_Elmer.log"\n'.format(
                        script_file,
                        Path(json_filename).relative_to(path),
                        python_executable)
                    )
                    file.write('echo "Simulation {}/{} Paraview"\n'.format(i + 1, len(json_filenames)))
                    file.write('{2} "{0}" "{1}" --only-paraview\n'.format(
                        script_file,
                        Path(json_filename).relative_to(path),
                        python_executable)
                    )

                    file.write('echo "Simulation {}/{} write results json"\n'.format(i + 1, len(json_filenames)))
                    file.write('{2} "{0}" "{1}" --write-project-results 2>&1 >> '\
                            '"{1}_write_project_results.log"\n'.format(
                        script_file,
                        Path(json_filename).relative_to(path),
                        python_executable)
                    )

                    file.write('echo "Simulation {}/{} write versions file"\n'.format(i + 1, len(json_filenames)))
                    file.write('{2} "{0}" "{1}" --write-versions-file\n'.format(
                        script_file,
                        Path(json_filename).relative_to(path),
                        python_executable)
                    )
                # change permission
                os.chmod(script_filename, os.stat(script_filename).st_mode | stat.S_IEXEC)
                if parallelize_workload:
                    main_file.write(' "./{}"'.format(
                        Path(script_filename).relative_to(path))
                    )
                else:
                    main_file.write('echo "Submitting the main script of simulation {}/{}"\n'
                            .format(i + 1, len(json_filenames)))
                    main_file.write('echo "--------------------------------------------"\n')
                    main_file.write('"./{}"\n'.format(Path(script_filename).relative_to(path)))

    # change permission
    os.chmod(main_script_filename, os.stat(main_script_filename).st_mode | stat.S_IEXEC)

    return main_script_filename


def export_elmer(simulations: Sequence[Simulation],
                 path: Path,
                 tool='capacitance',
                 linear_system_method='bicgstab',
                 p_element_order=3,
                 frequency=5,
                 file_prefix='simulation',
                 script_file='scripts/run.py',
                 mesh_size=None,
                 boundary_conditions=None,
                 workflow=None,
                 percent_error=0.005,
                 max_error_scale=2,
                 max_outlier_fraction=1e-3,
                 maximum_passes=1,
                 minimum_passes=1,
                 dielectric_surfaces=None,
                 is_axisymmetric=False,
                 skip_errors=False):
    """
    Exports an elmer simulation model to the simulation path.

    Args:

        simulations(list(Simulation)): list of all the simulations
        path(Path): Location where to output the simulation model
        tool(str): Available: "capacitance", "wave_equation" and "cross-section" (Default: capacitance)
        linear_system_method(str): Available: 'bicgstab', 'mg' (Default: bicgstab)
        p_element_order(int): polynomial order of p-elements (Default: 1)
        frequency: Units are in GHz. To set up multifrequency analysis, use list of numbers.
        file_prefix: File prefix of the script file to be created.
        script_file: Name of the script file to run.
        mesh_size(dict): Parameters to determine mesh element sizes
        boundary_conditions(dict): Parameters to determine boundary conditions
        workflow(dict): Parameters for simulation workflow
        percent_error(float): Stopping criterion in adaptive meshing.
        max_error_scale(float): Maximum element error, relative to percent_error, allowed in individual elements.
        max_outlier_fraction(float): Maximum fraction of outliers from the total number of elements
        maximum_passes(int): Maximum number of adaptive meshing iterations.
        minimum_passes(int): Minimum number of adaptive meshing iterations.
        dielectric_surfaces: Loss tangents for dielectric interfaces, thickness and permittivity should be specified in
            the simulation. The loss tangent is post-processed to the participation to get the quality factor.
            Default is None. Input is of the form::

                'substrate': {
                    'tan_delta_surf': 5e-7
                },
                'layerMA': {  # metal–vacuum
                    'tan_delta_surf': 0.001,  # loss tangent
                },
                'layerMS': { # metal–substrate
                    'tan_delta_surf': 0.001,
                },
                'layerSA': { # substrate–vacuum
                    'tan_delta_surf': 0.001,
                }
        is_axisymmetric(bool): Simulate with Axi Symmetric coordinates along :math:`y\\Big|_{x=0}` (Default: False)
        skip_errors(bool): Skip simulations that cause errors. (Default: False)

            .. warning::

               **Use this carefully**, some of your simulations might not make sense physically and
               you might end up wasting time on bad simulations.

    Returns:

        Path to exported script file.
    """
    parser = argparse.ArgumentParser()

    parser.add_argument('-q', "--quiet", action='store_true')
    args, _ = parser.parse_known_args()

    if args.quiet:
        workflow.update(
            {
                'run_gmsh_gui': False,  # For GMSH: if true, the mesh is shown after it is done
                                       # (for large meshes this can take a long time)
                'run_paraview': False,  # this is visual view of the results
            })

    write_commit_reference_file(path)
    copy_elmer_scripts_to_directory(path)
    json_filenames = []
    for simulation in simulations:
        try:
            json_filenames.append(
                export_elmer_json(
                    simulation=simulation,
                    path=path,
                    tool=tool,
                    linear_system_method=linear_system_method,
                    p_element_order=p_element_order,
                    frequency=frequency,
                    mesh_size=mesh_size,
                    boundary_conditions=boundary_conditions,
                    workflow=workflow,
                    percent_error=percent_error,
                    max_error_scale=max_error_scale,
                    max_outlier_fraction=max_outlier_fraction,
                    maximum_passes=maximum_passes,
                    minimum_passes=minimum_passes,
                    dielectric_surfaces=dielectric_surfaces,
                    is_axisymmetric=is_axisymmetric
                )
            )
        except (IndexError, ValueError, Exception) as e:  # pylint: disable=broad-except
            if skip_errors:
                logging.warning(
                    f'Simulation {simulation.name} skipped due to {e.args}. '\
                    'Some of your other simulations might not make sense geometrically. '\
                    'Disable `skip_errors` to see the full traceback.'
                )
            else:
                raise UserWarning(
                    'Generating simulation failed. You can discard the errors using `skip_errors` in `export_elmer`. '\
                    'Moreover, `skip_errors` enables visual inspection of failed and successful simulation '\
                    'geometry files.'
                ) from e

    return export_elmer_script(json_filenames, path, workflow, file_prefix=file_prefix, script_file=script_file)



# This code is part of KQCircuits
# Copyright (C) 2022 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).


from math import sqrt
from kqcircuits.pya_resolver import pya
from kqcircuits.util.parameters import Param, pdt
from kqcircuits.junctions.junction import Junction
from kqcircuits.util.symmetric_polygons import polygon_with_vsym


class ManhattanSingleJunction(Junction):
    """The PCell declaration for a Manhattan style single junction.
    """

    finger_overshoot = Param(pdt.TypeDouble, "Length of fingers after the junction.", 1.0, unit="μm")
    include_base_metal_gap = Param(pdt.TypeBoolean, "Include base metal gap layer.", True)
    include_base_metal_addition = Param(pdt.TypeBoolean, "Include base metal addition layer.", True)
    shadow_margin = Param(pdt.TypeDouble, "Shadow layer margin near the the pads.", 0.5, unit="μm")
    separate_junctions = Param(pdt.TypeBoolean, "Junctions to separate layer.", False)
    offset_compensation = Param(pdt.TypeDouble, "Junction lead offset from junction width", 0, unit="μm")
    mirror_offset = Param(pdt.TypeBoolean, "Move the junction lead offset to the other lead", False)
    finger_overlap = Param(pdt.TypeDouble, "Length of fingers inside the pads.", 1.0, unit="μm")
    height = Param(pdt.TypeDouble, "Height of the junction element.", 22.0, unit="μm")
    width = Param(pdt.TypeDouble, "Width of the junction element.", 22.0, unit="μm")
    pad_height = Param(pdt.TypeDouble, "Height of the junction pad.", 6.0, unit="μm")
    pad_width = Param(pdt.TypeDouble, "Width of the junction pad.", 12.0, unit="μm")
    pad_to_pad_separation = Param(pdt.TypeDouble, "Pad separation.", 6.0, unit="μm")
    x_offset = Param(pdt.TypeDouble, "Horizontal junction offset.", 0, unit="μm")
    pad_rounding_radius = Param(pdt.TypeDouble, "Rounding radius of the junction pad.", 0.5, unit="μm")

    def build(self):
        self.produce_manhattan_junction()

    def produce_manhattan_junction(self):

        # corner rounding parameters
        rounding_params = {
            "rinner": self.pad_rounding_radius,  # inner corner rounding radius
            "router": self.pad_rounding_radius,  # outer corner rounding radius
            "n": 64,  # number of point per rounded corner
        }

        junction_shapes_top = []
        junction_shapes_bottom = []
        shadow_shapes = []

        # create rounded bottom part
        y0 = (self.height / 2) - self.pad_height/2
        bp_pts_left = [
            pya.DPoint(-self.pad_width / 2, y0),
            pya.DPoint(-self.pad_width / 2, y0 + self.pad_height)
        ]
        bp_shape = pya.DTrans(0, False, 0,
                              -self.pad_height / 2 - self.pad_to_pad_separation/2) * polygon_with_vsym(bp_pts_left)
        self._round_corners_and_append(bp_shape, junction_shapes_bottom, rounding_params)

        bp_shadow_pts_left = [
            bp_pts_left[0] + pya.DPoint(-self.shadow_margin, -self.shadow_margin),
            bp_pts_left[1] + pya.DPoint(-self.shadow_margin, self.shadow_margin)
        ]
        bp_shadow_shape = pya.DTrans(0, False, 0,
                                     -self.pad_height / 2
                                     - self.pad_to_pad_separation / 2) * polygon_with_vsym(bp_shadow_pts_left)
        self._round_corners_and_append(bp_shadow_shape, shadow_shapes, rounding_params)

        # create rounded top part
        tp_shape = pya.DTrans(0, False, 0,
                              self.pad_height/2 + self.pad_to_pad_separation/2) * polygon_with_vsym(bp_pts_left)
        self._round_corners_and_append(tp_shape, junction_shapes_top, rounding_params)

        tp_shadow_shape = pya.DTrans(0, False, 0,
                                     self.pad_height/2 +
                                     self.pad_to_pad_separation/2) * polygon_with_vsym(bp_shadow_pts_left)
        self._round_corners_and_append(tp_shadow_shape, shadow_shapes, rounding_params)

        # create rectangular junction-support structures and junctions
        self._make_junction(pya.DPoint(0, self.height / 2 + 2.8), self.height / 2 - 5, 0)
        self._add_shapes(junction_shapes_bottom, "SIS_junction")
        self._add_shapes(junction_shapes_top, "SIS_junction")
        self._add_shapes(shadow_shapes, "SIS_shadow")
        self._produce_ground_metal_shapes()
        self._produce_ground_grid_avoidance()
        self._add_refpoints()

    def _make_junction(self, top_corner, b_corner_y, finger_margin=0):
        """Create junction fingers and add them to some SIS layer.
        Choose 'SIS_junction' layer by default but 'SIS_junction_2' if ``separate_junctions`` is True.
        """
        jx = top_corner.x - (top_corner.y - b_corner_y) / 2
        jy = (top_corner.y + b_corner_y) / 2
        ddb = self.junction_width * sqrt(0.5)
        ddt = self.junction_width * sqrt(0.5)
        if self.mirror_offset:
            ddt += self.offset_compensation * sqrt(0.5)
        else:
            ddb += self.offset_compensation * sqrt(0.5)
        fo = self.finger_overshoot * sqrt(0.5) - 1.1
        pl = self.finger_overlap * sqrt(0.5) + 0.2  # plus length to connect despite of rounding

        def finger_points(size):
            return [
                pya.DPoint(top_corner.x + pl, top_corner.y + size + pl),
                pya.DPoint(top_corner.x + size + pl, top_corner.y + pl),
                pya.DPoint(jx - fo, jy - fo - size),
                pya.DPoint(jx - fo - size, jy - fo),
            ]

        finger_bottom = pya.DTrans(-jx, -jy + self.x_offset) * pya.DPolygon(finger_points(ddb))
        finger_top = pya.DTrans(-jx + self.x_offset, -jy) * pya.DPolygon(finger_points(ddt))

        junction_shapes = [(pya.DTrans(jx - finger_margin, jy) * finger_top).to_itype(self.layout.dbu),
                           (pya.DTrans(0, False, jx-2*top_corner.x, jy) * finger_top).to_itype(self.layout.dbu),
                           (pya.DTrans(3, False, jx - finger_margin, jy+2.2) * finger_bottom).to_itype(self.layout.dbu),
                           (pya.DTrans(3, False, jx-2*top_corner.x, jy+2.2) * finger_bottom).to_itype(self.layout.dbu)]

        junction_region = pya.Region(junction_shapes).merged()
        layer_name = "SIS_junction_2" if self.separate_junctions else "SIS_junction"
        self.cell.shapes(self.get_layer(layer_name)).insert(junction_region)

        # place refpoint at the middle of the junctions
        self.refpoints["c"] = pya.DPoint(jx + 1.1 - finger_margin, jy + 1.1)

    def _add_shapes(self, shapes, layer):
        """Merge shapes into a region and add it to layer."""
        region = pya.Region(shapes).merged()
        self.cell.shapes(self.get_layer(layer)).insert(region)

    def _add_refpoints(self):
        """Adds the "origin_squid" refpoint and port "common"."""
        self.refpoints["origin_squid"] = pya.DPoint(0, 0)
        self.add_port("common", pya.DPoint(0, self.height))

    def _produce_ground_metal_shapes(self):
        """Produces hardcoded shapes in metal gap and metal addition layers."""
        # metal additions bottom
        x0 = - self.a / 2
        y0 = self.height / 2
        bottom_pts = [
            pya.DPoint(x0 + 2, y0 - 7),
            pya.DPoint(x0 + 2, y0 - 5),
            pya.DPoint(x0 + 3, y0 - 5),
            pya.DPoint(x0 + 3, y0 - 4),
            pya.DPoint(x0, y0 - 4),
            pya.DPoint(x0, 0)
        ]
        if self.include_base_metal_addition:
            shape = polygon_with_vsym(bottom_pts)
            self.cell.shapes(self.get_layer("base_metal_addition")).insert(shape)
            # metal additions top
            top_pts = [
                pya.DPoint(x0 + 2, y0 + 7),
                pya.DPoint(x0 + 2, y0 + 5),
                pya.DPoint(x0 + 3, y0 + 5),
                pya.DPoint(x0 + 3, y0 + 4),
                pya.DPoint(x0, y0 + 4),
                pya.DPoint(x0, self.height),
            ]

            shape = polygon_with_vsym(top_pts)
            self.cell.shapes(self.get_layer("base_metal_addition")).insert(shape)
        # metal gap
        if self.include_base_metal_gap:
            if self.include_base_metal_addition:
                pts = bottom_pts + [pya.DPoint(-self.width / 2, 0), pya.DPoint(-self.width / 2, self.height)] \
                      + top_pts[::-1]
            else:
                pts = [pya.DPoint(-self.width / 2, 0), pya.DPoint(-self.width / 2, self.height)]
            shape = polygon_with_vsym(pts)
            self.cell.shapes(self.get_layer("base_metal_gap_wo_grid")).insert(shape)

    def _produce_ground_grid_avoidance(self):
        """Add ground grid avoidance."""
        w = self.cell.dbbox().width()
        h = self.cell.dbbox().height()
        protection = pya.DBox(-w / 2 - self.margin, - self.margin, w / 2 + self.margin, h + self.margin)
        self.cell.shapes(self.get_layer("ground_grid_avoidance")).insert(protection)

    def _round_corners_and_append(self, polygon, polygon_list, rounding_params):
        """Rounds the corners of the polygon, converts it to integer coordinates, and adds it to the polygon list."""
        polygon = polygon.round_corners(rounding_params["rinner"], rounding_params["router"], rounding_params["n"])
        polygon_list.append(polygon.to_itype(self.layout.dbu))



# This code is part of KQCircuits
# Copyright (C) 2021 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).

from importlib import util, import_module

# Record dependencies in setup.py too
kqc_python_dependencies = {
    "numpy": "numpy>=1.16",
    "scipy": "scipy>=1.2",
    "tqdm": "tqdm>=4.61",
    "networkx": "networkx>=2.7",
}


def install_kqc_dependencies():
    """Check KQCircuits' dependencies and install/upgrade if missing.

    This is *only* for KLayout. Stand-alone mode needs manual pip install, preferably in a venv.
    This function should run only once at KLayout startup.
    """
    # pylint: disable=import-outside-toplevel

    from kqcircuits.pya_resolver import pya
    # Skip installation in stand-alone python package mode
    if not hasattr(pya, 'MessageBox'):
        return

    missing_pkgs = []

    for pkg, pkg_ver in kqc_python_dependencies.items():
        if util.find_spec(pkg) is None:
            missing_pkgs.append(pkg)
        else:
            mod = import_module(pkg)
            v_inst = tuple(map(int, (mod.__version__.split("."))))
            v_reqd = tuple(map(int, (pkg_ver.rsplit("=", maxsplit=1)[-1].split("."))))
            if v_reqd > v_inst:
                missing_pkgs.append(pkg)

    if not missing_pkgs:
        return

    # Install missing modules inside KLayout.
    from pip import __main__
    if hasattr(__main__, "_main"):
        main = __main__._main
    else:
        from pip._internal.cli.main import main

    ask = pya.MessageBox.warning("Install packages?", "Install missing packages using 'pip': " +
                                 ", ".join(missing_pkgs), pya.MessageBox.Yes + pya.MessageBox.No)
    if ask == pya.MessageBox.Yes:
        main(['install'] + [kqc_python_dependencies[pkg] for pkg in missing_pkgs])



# This code is part of KQCircuits
# Copyright (C) 2021 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).


import math

from kqcircuits.elements.element import Element
from kqcircuits.elements.fluxlines.fluxline import Fluxline
from kqcircuits.pya_resolver import pya
from kqcircuits.util.parameters import Param, pdt, add_parameters_from
from kqcircuits.junctions.junction import Junction
from kqcircuits.junctions.squid import Squid
from kqcircuits.junctions.sim import Sim


@add_parameters_from(Fluxline, "fluxline_gap_width", "fluxline_type", "fluxline_parameters", "_fluxline_parameters")
@add_parameters_from(Squid, "junction_width", "loop_area", "junction_type",
                     "junction_parameters", "_junction_parameters")
@add_parameters_from(Sim, "junction_total_length")
class Qubit(Element):
    """Base class for qubit objects without actual produce function.

    Collection of shared sub routines for shared parameters and producing shared aspects of qubit geometry including

    * possible fluxlines
    * e-beam layers for SQUIDs
    * SQUID name parameter
    """

    LIBRARY_NAME = "Qubit Library"
    LIBRARY_DESCRIPTION = "Library for qubits."
    LIBRARY_PATH = "qubits"

    mirror_squid =  Param(pdt.TypeBoolean, "Mirror SQUID by its Y axis", False)

    def coerce_parameters_impl(self):
        self.sync_parameters(Fluxline)
        self.sync_parameters(Junction)

    def produce_squid(self, transf, only_arms=False, **parameters):
        """Produces the squid.

        Creates the squid cell and inserts it with the given transformation as a subcell. Also inserts the squid parts
        in "base_metal_gap_wo_grid"-layer to "base_metal_gap_for_EBL"-layer. It also returns a ``right_side`` refpoint,
        calculated from base_metal_gap_wo_grid layer's bounding box to help with arm_length calculation in
        JunctionTestPads.

        Args:
            transf (DCplxTrans): squid transformation
            parameters: other parameters for the squid
            only_arms: Boolean argument that allows to choose whether to create the arms and the squid device or only
                       the arms

        Returns:
            (dict): Relative refpoints for the squid

        """
        cell = self.add_element(Squid, junction_type=self.junction_type, **parameters)
        refpoints_rel = self.get_refpoints(cell)
        mwidth = cell.dbbox_per_layer(self.get_layer("base_metal_gap_wo_grid")).width()
        if mwidth > 0.0:
            refpoints_rel['right_side'] = pya.DPoint(mwidth / 2, 0.0)
        squid_transf = transf * pya.DTrans.M90 if self.mirror_squid else transf

        if "squid_index" in parameters:
            s_index = int(parameters.pop('squid_index'))
            inst, _ = self.insert_cell(cell, squid_transf, inst_name=f"squid_{s_index}")
            inst.set_property("squid_index", s_index)
        else:
            inst, _ = self.insert_cell(cell, squid_transf, inst_name="squid")

        if only_arms:
            # add squid metal etch and unetch shapes to qubit and erase instance of squid
            for layer_name in ["base_metal_gap_wo_grid", "base_metal_addition"]:
                region = pya.Region(cell.shapes(self.get_layer(layer_name)))
                region.transform(inst.cplx_trans)
                self.cell.shapes(self.get_layer(layer_name)).insert(region)
            self.cell.erase(inst)
        else:
            # add parts of qubit to the layer needed for EBL
            region = pya.Region(cell.shapes(self.get_layer("base_metal_gap_wo_grid")))
            region.transform(inst.cplx_trans)
            self.cell.shapes(self.get_layer("base_metal_gap_for_EBL")).insert(region)

        return refpoints_rel

    def produce_fluxline(self, rot=0, trans=pya.DVector(), **parameters):
        """Produces the fluxline.

        Creates the fluxline cell and inserts it as a subcell. The "flux" and "flux_corner" ports
        are made available for the qubit. By default, fluxlines align their "origin_fluxline" refpoint to
        "origin_squid" refpoint in the direction of "port_common". However, the user might tweak the alignment direction
        by using the argument rot and the relative position by an extra pya.DVector(x, y) allowing to tune the position
        to achieve the desired design parameters.

        Args:
            rot: Extra rotation of the fluxline, in degrees
            trans (DVector): fluxline x/y translation
            parameters: parameters for the fluxline to overwrite default and subclass parameters
        """

        if self.fluxline_type == "none":
            return

        cell = self.add_element(Fluxline, **parameters)

        refpoints_so_far = self.get_refpoints(self.cell)
        squid_edge = refpoints_so_far["origin_squid"]
        a = (squid_edge - refpoints_so_far['port_common'])
        rotation = math.atan2(a.y, a.x) / math.pi * 180 + 90
        total_transformation = pya.DCplxTrans(1, rotation + rot, False, squid_edge - self.refpoints["base"] + trans)

        cell_inst, _ = self.insert_cell(cell, total_transformation)
        self.copy_port("flux", cell_inst)



# This code is part of KQCircuits
# Copyright (C) 2021 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).

def count_instances_in_cell(cell, pcell_class):
    """Returns the number of pcell instances of type `pcell_class` in cell.

    The instances are counted from the entire hierarchy below cell, not only direct child instances. Also pcells with
    type derived from `pcell_class` are counted.

    Args:
        cell: cell from which the instances are counted
        pcell_class: pcell class of the instances

    Returns:
        The number of instances below `cell` for which `isinstance(inst.cell.pcell_declaration(), pcell_class) == True`.
    """
    n = 0
    if isinstance(cell.pcell_declaration(), pcell_class):
        n += 1
    for inst in cell.each_inst():
        # workaround for getting the cell due to KLayout bug, see
        # https://www.klayout.de/forum/discussion/1191/cell-shapes-cannot-call-non-const-method-on-a-const-reference
        # TODO: replace by `inst_cell = inst.cell` once KLayout bug is fixed
        inst_cell = cell.layout().cell(inst.cell_index)
        n += count_instances_in_cell(inst_cell, pcell_class)
    return n



# This code is part of KQCircuits
# Copyright (C) 2021 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).

import logging
from itertools import product
from pathlib import Path

from kqcircuits.simulations.export.util import export_layers


def export_simulation_oas(simulations, path: Path, file_prefix='simulation'):
    """
    Write single OASIS file containing all simulations in list.
    """
    unique_layouts = {simulation.layout for simulation in simulations}
    if len(unique_layouts) != 1:
        raise ValueError("Cannot write batch OASIS file since not all simulations are on the same layout.")

    cells = [simulation.cell for simulation in simulations]
    oas_filename = str(path.joinpath(file_prefix + '.oas'))
    export_layers(oas_filename, simulations[0].layout, cells,
                  output_format='OASIS',
                  layers=None)
    return oas_filename


def sweep_simulation(layout, sim_class, sim_parameters, sweeps):
    """Create simulation sweep by varying one parameter at time. Return list of simulations."""
    simulations = []
    lengths = [len(l) for l in sweeps.values()]
    logging.info(f'Added simulations: {" + ".join([str(l) for l in lengths])} = {sum(lengths)}')
    for param in sweeps:
        for value in sweeps[param]:
            parameters = {**sim_parameters, param: value,
                          'name': '{}_{}_{}'.format(
                            sim_parameters['name'],
                            param,
                            str(value)
                          )}
            simulations.append(sim_class(layout, **parameters))
    return simulations


def cross_sweep_simulation(layout, sim_class, sim_parameters, sweeps):
    """Create simulation sweep by cross varying all parameters. Return list of simulations."""
    simulations = []
    keys = list(sweeps)
    sets = [list(prod) for prod in product(*sweeps.values())]
    logging.info(f'Added simulations: {" * ".join([str(len(l)) for l in sweeps.values()])} = {len(sets)}')
    for values in sets:
        parameters = {**sim_parameters}
        for i, key in enumerate(keys):
            parameters[key] = values[i]
        parameters['name'] = sim_parameters['name'] + '_' \
            + '_'.join([str(value) for value in values])
        simulations.append(sim_class(layout, **parameters))
    return simulations
